/* classifiers.h
 * Michael Sparks (mespar1@iastate.edu)
 * Last modified : 11 June 2007
 *
 * Contains classifier-related functions for use in PASIF
 * and/or the learnclassifier utility.
 *
 * Copyright (c) 2007 Michael E Sparks
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef CLASSIFIERS_H
#define CLASSIFIERS_H

#include "immpractical.h"

/* Data structures for handling machine learning classifiers. ****************/

#define MAXERROFWEIGHTS    10E-4 /* Target sum-of-squares error for  *
                                  * learned neuron weights.          */
#define AVAILCLASSIFIERCT      3 /* Number of classification methods *
                                  * available to PASIF.              */
enum { /* Which dimension corresponds to which classifier? *
        * It is critical that ORGOBJFCTX be the last elt   */
  LINEARUNITX=0,
  SIGMOIDUNITX,
  ORGOBJFCTX
};
/* English descriptions of the available classifiers */
static const char classdesc[AVAILCLASSIFIERCT][MAXLOCSTRING] = {
  "Unthresholded Perceptron (Linear unit)",
  "Sigmoid Unit",
  "Original objective function"
};
static const char classifiertags[AVAILCLASSIFIERCT-1][MAXLOCSTRING] = {
  "linunit",
  "sigunit"
};

/* Structure representing descriptive feature vector of a spliced isoform */
typedef struct {
  double
    donmeanave, /* Mean average of donor site posterior probs     */
    accmeanave, /* Mean average of acceptor site posterior probs  */
    donmeanrho, /* Mean average of donor site rho-values          */
    accmeanrho, /* Mean average of acceptor site rho-values       */
    zetacomp,   /* Gene struct's composite zeta-value based score */
    codprob;    /* Posterior prob of induced  coding sequence     */
} featvecT;

/* Structure for bundling data needed to produce a featvecT object */
typedef struct {
  int *locus;           /* int translation of locus       */
  int locuslength;      /* stores length of locus array   */
  double *donscores;    /* decoration of locus with a     *
                         * posteriori donor site probs    */
  double *accscores;    /* decoration of locus with a     *
                         * posteriori acceptor site probs */
  double *rho_vals_don; /* rho-values under donor id      */
  double *rho_vals_acc; /* rho-values under acceptor id   */
  int MMalgotype;       /* Code for Markov model variant  */
  foprobT *fixordprobs; /* FO parameters                  */
  immprobT *immprobs;   /* TDDI,BUDI,CHI2 parameters      */
  dmprobT *dmprobs;     /* DMMM parameters                */
  int cloudtype;        /* var- or range-based [DP]MMM?   */
} data4makefeatvecT;

/* Data structure for a trained perceptron classifier */
typedef struct {
  /* Weights for descriptive attributes */
  double
    wzero, /* Offset weight */
    wdonmeanave,
    waccmeanave,
    wdonmeanrho,
    waccmeanrho,
    wzetacomp,
    wcodprob;
} LTUparmT;

/* Function prototypes *******************************************************/

/* Wraps the process of adjusting an LTU's weight array */
void update_weight_vector(
  featvecT features,  /* Feature vector of instance     */
  LTUparmT *weights,  /* Weight vector to modify        */
  double target,      /* Records true class of instance */
  double output,      /* Output of activation unit      */
  int classifiercode, /* see classifiers.h              */
  double learnrate    /* Learning rate                  */
);

/* Add components of vector arguments */
LTUparmT add_weight_vectors(
  LTUparmT ltu1,
  LTUparmT ltu2
);

/* Assay an input instance using a trained neuron */
double eval_activation_unit(
  featvecT features,     /* Feature vector of instance         */
  LTUparmT trainedmodel, /* Learned weights for the LTU        */
  int *classification,   /* Records (Boolean) result of        *
                          * classification                     */
  double *fctofdotprod,  /* Stores result of activation fct    *
                          * on dot product of weight/feat vecs */
  int classifiercode     /* see classifiers.h                  */
);

/* Function to import a trained LTUparmT object */
void import_trained_perceptron(char *filename,LTUparmT *parms);

/* Implements PASIF's original, very primitive objective function */
double orgobjfct(
  featvecT features,   /* Feature vector of isoform               */
  int *classification, /* Records result of classification        */
  double *score        /* Records numeric score of classification */
);

#endif
