/* immpractical.h
 * Michael E Sparks (mespar1@iastate.edu)
 * Last modified : 11 June 2007
 *
 * Header file contains various definitions acting more or less
 * globally on code for the immpractical library.
 *
 * Copyright (C) 2005,2006,2007 Michael E Sparks
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef IMMPRACTICAL_H
#define IMMPRACTICAL_H

/* Parametric String Literals ************************************************/

#ifndef TRUE
#define TRUE                          1 /* Truth value                      */
#endif
#ifndef FALSE
#define FALSE                         0 /* False value                      */
#endif

#define ALFSIZE                       4 /* A,C,G,[T|U] ... no ambiguity     *
                                         * permitted since we're building a *
                                         * model, afterall!                 */
#define MAXDIFF                   10E-6 /* Max deviation for float compare  */
#define MAXLOCSTRING               1024 /* Max length for any local strings */
#define MAXORDER                      5 /* [0..5] (pretext/history lengths) */
#ifdef THREEPERIODIC
#ifdef SHADOWSTRAND
#define NUMMODELS                     7 /* Model 0 (C g t),  (forward)      *
                                         *       1 (c G t),  (forward)      *
                                         *       2 (c g T),  (forward)      *
                                         *       3 (A c g),  (reverse)      *
                                         *       4 (a C g),  (reverse)      *
                                         *       5 (a c G),  (reverse)      *
                                         *       6 noncoding (intron)       */
#define NONCODING                     6 /* Noncoding model index            */
#else
#define NUMMODELS                     4 /* Model 0 (C g t),  (forward)      *
                                         *       1 (c G t),  (forward)      *
                                         *       2 (c g T),  (forward)      *
                                         *       3 noncoding (intron)       */
#define NONCODING                     3 /* Noncoding model index            */
#endif
#else
#define NUMMODELS                     2 /* Coding (0) and Noncoding (1)     */
#define NONCODING                     1 /* Noncoding model index            */
#endif

/* Algorithm specifications **************************************************/

#define NUMALGOS   6 /* Number of training algorithms  *
                      * the library (ideally) supports */

/* Which dimension corresponds to which algorithm? */
enum {
  FIXORDX=0,
  TDDIX,
  BUDIX,
  CHI2X,
  DMMMX,
  PMMMX
};

/* The following char strings describe the training/testing *
 * algorithms currently implented in the library.           */
static const char algodesc[NUMALGOS][MAXLOCSTRING] = {
  "fixed-order (pseudocount smoothed)",
  "top-down deleted interpolation",
  "bottom-up deleted interpolation",
  "chi-square",
  "dynamically modulating",
  "partially modulating mixed model"
};

/* The following tags are appended to the binary parameter file, *
 * indicating from which statistical model the paramters were    *
 * actually generated.  Please ensure indices correspond to      *
 * those of algodesc!                                            */
static const char algotags[NUMALGOS][MAXLOCSTRING] = {
  "FIXORD",
  "TDDI",
  "BUDI",
  "CHI2",
  "DMMM",
  "PMMM"
};

/* Indices for DMMM methods */
enum {
  VARIANCEBASED=0,
  RANGEBASED
};

/* Structure definitions *****************************************************/

typedef struct {
  /* These arrays will store oligo frequencies, mononucleotides *
   * up through hexanucleotides.  These will be used to derive  *
   * max likelihood probabilities.  The NUMMODELS dimension may *
   * at first look unnecessary, but it IS required for counting *
   * three-periodically.                                        */
  int
    count1[NUMMODELS][ALFSIZE],
    count2[NUMMODELS][ALFSIZE][ALFSIZE],
    count3[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE],
    count4[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE],
    count5[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE],
    count6[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE];
} countarrayT;

typedef struct {
  /* Arrays to store the relative frequencies (unsmoothed *
   * maximum likelihood estimates) of oligomers.          */
  double
    MLhis0[ALFSIZE],
    MLhis1[ALFSIZE][ALFSIZE],
    MLhis2[ALFSIZE][ALFSIZE][ALFSIZE],
    MLhis3[ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE],
    MLhis4[ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE],
    MLhis5[ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE];
} relfreqsT;

typedef struct {
  /* Arrays to store incrementally smoothed ML estimates in *
   * the bottom-up training procedure. The first dimension  *
   * keys the "innerorder" (i=k..0), and other dimensions   *
   * key the oligos being trained for.  Each array          *
   * corresponds to a given "order" (k).                    */ 
  double
    Psuphis0[1][ALFSIZE],
    Psuphis1[2][ALFSIZE][ALFSIZE],
    Psuphis2[3][ALFSIZE][ALFSIZE][ALFSIZE],
    Psuphis3[4][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE],
    Psuphis4[5][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE],
    Psuphis5[6][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE];
} PsupiT;

typedef struct {
  /* Arrays store final, possibly smoothed probabilities of *
   * oligomers of length 1 to MAXORDER + 1 for fast lookup  *
   * by the IMMprob function.                               */
  double
    prob1[NUMMODELS][ALFSIZE],
    prob2[NUMMODELS][ALFSIZE][ALFSIZE],
    prob3[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE],
    prob4[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE],
    prob5[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE],
    prob6[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE];
} immprobT;

/* Structure stores fixed-order Markov model- *
 * derived probabilities for oligomers.       */
typedef struct {
  /* MAXORDER can vary from 0..5, inclusive. I use this same      *
   * data structure for any of these orders: for MAXORDER == 5    *
   * I key into the data structure using all 6 nucleotides, while *
   * for shorter oligos I "blank out" higher dimensional indices  *
   * with 0's. It's the programmer's responsibility to do this    *
   * consistently when referencing these data.                    */  
  double
    FOprob[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE][ALFSIZE];
} foprobT;

/* Structure stores arithmetic mean, standard deviation, min, and   *
 * max of hexamer probability estimates derived from partitioning   *
 * the training data according to some scheme, e.g., random splits, *
 * GC-content-based splits, etc.                                    */ 
typedef struct {
  /* Element names are self-explanatory */
  double
    mean[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE]
                   [ALFSIZE][ALFSIZE][ALFSIZE],
    stddev[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE]
                     [ALFSIZE][ALFSIZE][ALFSIZE],
    min[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE]
                  [ALFSIZE][ALFSIZE][ALFSIZE],
    max[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE]
                  [ALFSIZE][ALFSIZE][ALFSIZE],
    muless1sigma[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE]
                           [ALFSIZE][ALFSIZE][ALFSIZE],
    muplus1sigma[NUMMODELS][ALFSIZE][ALFSIZE][ALFSIZE]
                           [ALFSIZE][ALFSIZE][ALFSIZE];
} dmprobT;

/* Public function prototypes ************************************************/

/* Function to import a trained foprobT     *
 * object (a binary file) into main memory. */
void import_fo_probs(char *filename,foprobT *probs);

/* Function to import a trained immprobT  *
 * object (binary file) into main memory. */
void import_imm_probs(char *filename,immprobT *probs);

/* Function to import a trained dmprobT     *
 * object (a binary file) into main memory. */
void import_dm_probs(char *filename,dmprobT *probs);

/* Function to return log-likelihood of a test input sequence, *
 * under the model specified using the model argument.         *
 * If threeper is FALSE, the                                   *
 * function returns the non-periodic log-probability of the    *
 * test sequence.  If non-FALSE, this function handles those   *
 * cases where model is an element of {0,1,2} if shadow is     *
 * FALSE, or {0,1,2,3,4,5} if shadow is non-FALSE; it returns  *
 * the probability of the input sequence given its first base  *
 * occurs in the phase specified by the model argument.        *
 * (see immpractical.h for phase nomenclature).                */
long double FOprob(int *nt,int length,foprobT *probs,
                   int model,int threeper,int shadow);

/* Function to return log-likelihood of a test input sequence, *
 * under the model specified using the model argument.         *
 * If threeper is FALSE, the                                   *
 * function returns the non-periodic log-probability of the    *
 * test sequence.  If non-FALSE, this function handles those   *
 * cases where model is an element of {0,1,2} if shadow is     *
 * FALSE, or {0,1,2,3,4,5} if shadow is non-FALSE; it returns  *
 * the probability of the input sequence given its first base  *
 * occurs in the phase specified by the model argument.        *
 * (see immpractical.h for phase nomenclature).                */
long double IMMprob(int *nt,int length,immprobT *probs,
                    int model,int threeper,int shadow);

/* Function to return log-likelihood of a test input sequence  *
 * under the model specified using the model argument.         *
 * If threeper is FALSE, the                                   *
 * function returns the non-periodic log-probability of the    *
 * test sequence.  If non-FALSE, this function handles those   *
 * cases where model is an element of {0,1,2} if shadow is     *
 * FALSE, or {0,1,2,3,4,5} if shadow is non-FALSE; it returns  *
 * the probability of the input sequence given its first base  *
 * occurs in the phase specified by the model argument.        *
 * (see immpractical.h for phase nomenclature).                */
long double DMMMprob(int *nt,int length,dmprobT *probs,
                     int cloudtype,int model,int threeper,
                     int shadow);

#endif
