/* pasif_utils.h
 * Michael Sparks (mespar1@iastate.edu)
 * Last modified : 13 June 2007
 *
 * Contains functions used by PASIF for conducting (local heuristic)
 * searches through candidate gene structure space.
 *
 * Copyright (c) 2006,2007 Michael E Sparks
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef PASIF_UTILS_H
#define PASIF_UTILS_H

#include "classifiers.h"
#include "immpractical.h"

#define XMLOUTSTREAM stdout /* Stream to print XML results to. */

/* Information used by the parallel PASIF system */
#ifdef MPI
#define MASTER           0 /* Master node's rank in communicator. */
#define MESGTAG        155 /* Message tag, hopefully unique. :|   */
#define MINNODESTOFARM 200 /* Do not distribute a workload of     *
                            * fewer nodes than this.              */
#endif

#define DEFAULTCOMPZETA 1E-6 /* Composite zeta-value based score for *
                              * structures without a splice site w/  *
                              * probability .GT. DEFAULTSSPROB       */

/* The following is configuration information that seems *
 * much better to keep hidden from the end user.         */
#define SPLICEUPSTREAMLIM   5 /* 5' extent of splice site in local *
                               * structure space                   */
#define SPLICEDOWNSTREAMLIM 5 /* 3' extent of splice site in local *
                               * structure space                   */

#define TRANSUPSTREAMLIM    3 /* 5' extent of translation start site *
                               * in local structure space            */
#define TRANSDOWNSTREAMLIM  3 /* 3' extent of translation start site *
                               * in local structure space            */

#define MAXTIMESTOSAMPLEEXONCT 25 /* Limit on attempts to pick a reasonable *
                                   * value for the exonct variable          */
#define MAXTIMESTOSHAPESEED    50 /* Limit on attempts to induce a seed on  *
                                   * a locus using exonct exons             */

/* Data structures for storing general configuration information. *
 * If I lose somebody here, things should be clearer by looking   *
 * at the implementation of the parsegeneralinfo function.        */
#define EXINTLENBOUNDS 4 /* upper/lower bounds on permissable     *
                          * exon and intron lengths of solutions, *
                          * this just states how many limits we   *
                          * will consider.                        */
enum { /* For keying into lengthlimits array */
  MININTRONLENX=0,
  MAXINTRONLENX,
  MINEXONLENX,
  MAXEXONLENX
};
static const char lenlimittags[EXINTLENBOUNDS][15] = {
  "MININTRONLEN", /* The meanings of these names should be clear? */
  "MAXINTRONLEN",
  "MINEXONLEN",
  "MAXEXONLEN"
};
int lengthlimits[EXINTLENBOUNDS]; /* Array to store these bounds */
#define RESTARTLIMITX "RESTARTLIMIT" /* ID for said option */
int restartlimit; /* Max seed-and-refine cycles to execute */
#define TARGETISOFORMCTX "TARGETISOFORMCT"
int targetisoformct; /* Target number of unique AS isoforms */
#define MAXEXONCTFULLEXPANDX "MAXEXONCTFULLEXPAND" /* ID for said option  */
/* Maximum number of exons in a putative isoform for which exhaustive     *
 * enumeration of child nodes is feasible, i.e., enumeration of up to     *
 * (SPLICEUPSTREAMLIM+SPLICEDOWNSTREAMLIM+1)**((maxexonctfullexpand-1)*2) *
 * many (possibly) unique child nodes at each stage of expansion.         */
int maxexonctfullexpand;
#define MAXSOLNDEPTHX "MAXSOLNDEPTH" /* ID for said option */
/* Depth-limit for hill-climbing search.  If a local optimum has not *
 * been found by this depth, then the search just returns the best   *
 * solution it has already found.                                    */
int maxsolndepth;

/* Structure for representing individual exons in a genomic sequences */
typedef struct exoncoors {
  int start, /* exon start position (1st exon base) */
      stop;  /* exon stop position (final exon base) */
  struct exoncoors *nextelt; /* Pointer to next element in list */
} exoncoorsT;

/* Structure for representing solution space *
 * range covered by a node expansion         */
typedef struct exonterminusrange {
  int upstreamlim,   /* upstream bound for exon terminus */
      downstreamlim; /* downstream bound for exon terminus */
  struct exonterminusrange *nextelt; /* Pointer to next element in list */
} exonterminusrangeT;

/* Function prototypes *******************************************************/

/* This is the main function responsible for coordinating the *
 * process of predicting alternative splice isoforms.         */
void makeASforms(
  int *locus,           /* int translation of locus       */
  int locuslength,      /* stores length of locus array   */
  char *locusid,        /* description of locus           */
  double *exonctdist,   /* distribution of exon counts    *
                         * in known transcript isoforms   */
  int exonctdistsize,   /* largest index of exonctdist    */
  /* Parameters for pre-computed splice site data         */
  double *donscores,    /* decoration of locus with       *
                         * posterior donor site probs     */
  double *accscores,    /* decoration of locus with       *
                         * posterior acceptor site probs  */
  double *rho_vals_don, /* rho-values under donor id      */
  double *rho_vals_acc, /* rho-values under acceptor id   */
  /* Parameters for scoring induced coding sequence       */
  int MMalgotype,       /* Code for Markov model variant  */
  foprobT *fixordprobs, /* FO parameters                  */
  immprobT *immprobs,   /* TDDI,BUDI,CHI2 parameters      */
  dmprobT *dmprobs,     /* DMMM parameters                */
  int cloudtype,        /* var- or range-based [DP]MMM?   */
  /* Parameters for instance classifier function          */
  int classifiercode,   /* Index of classifier algorithm  */
  LTUparmT
    *classifierparms    /* Trained classifier object      */
);

#endif
