/* PASIF.c
 * Michael Sparks (mespar1@iastate.edu)
 * Last modified : 14 June 2007
 *
 * Main implementation file for PASIF, a program for
 * Prediction of Alternative Spliced IsoForms.
 *
 * Copyright (c) 2006,2007 Michael E Sparks
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* Include Statements */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "classifiers.h"
#include "confparsing.h"
#include "errors.h"
#include "immpractical.h"
#include "pasif_utils.h"
#include "sequence_parse.h"
#include "splice_utils.h"
#ifdef MPI
#include "mpi.h"
#endif

#define MINARGS 9

#define BADUSE {\
  fprintf(stderr,"\a\n\
Conceptual Usage:\n\n\
  %s loci.fas exonctdist.dat generalconf.dat foo.BSSMparms \\\n\
    classifieralgocode foo.classifierparms MMalgocode foo.MMparms\n\n\
  Where classifieralgocode is an integer selected from\n",argv[0]);\
  for(i=0;i<AVAILCLASSIFIERCT;++i)\
    fprintf(stderr,"    (%i) %s\n",i,classdesc[i]);\
  fprintf(stderr,"\n  MMalgocode is an integer selected from\n");\
  for(i=0;i<NUMALGOS;++i)\
    fprintf(stderr,"    (%i) %s\n",i,algodesc[i]);\
  fprintf(stderr,"\n");\
}

#define DMUSAGE "\a\n\
USAGE w/ DMMM parameters:\n\n\
  %s ... %i cloud_type bar.%s\n\n\
  Where cloud_type = %i -> variance-based cloud\n\
                   = %i -> range-based cloud\n\n"

#define PMUSAGE "\a\n\
USAGE w/ PMMM:\n\n\
  %s ... %i cloud_type bar1.%s bar2.%s\n\n\
  Where cloud_type = %i -> variance-based cloud\n\
                   = %i -> range-based cloud\n\n"

/* Main Application */
int main(int argc, char *argv[]) {
  Bssmparm *bssmparms=NULL;    /* Stores Bssmparm object        */
  char *charloc=NULL,          /* genomic locus to be processed */   
       locusid[MAXDESCLINE+1], /* description of locus, +1 for  *
                                * terminating '\0'.             */
       *test=NULL;             /* Verifies MM file extension    */
  dmprobT *dmprobs=NULL;       /* Stores dmprobT object         */
  double *donscores=NULL,      /* don site post probabilities   */
         *accscores=NULL,      /* acc site post probabilities   */
         *rho_vals_don=NULL,   /* rho-values under donor id     */
         *rho_vals_acc=NULL,   /* rho-values under acceptor id  */
         *exonctdist=NULL;     /* distribution of exon counts   *
                                * in known transcript isoforms  */
  FILE *genfile=NULL;          /* genomic input file            */
  foprobT *fixordprobs=NULL;   /* Stores foprobT object         */
  immprobT *immprobs=NULL;     /* Stores immprobT object        */
  int *locus=NULL,             /* int translation of sequence   */
      locuslength=0,           /* stores length of locus array  */
      MMalgocode,              /* records code of Markov model  *
                                * algorithm to be used          */
      cloudtype=0,             /* records code of cloud type to *
                                * use with DMMM                 */
      exonctdistsize,          /* largest index of exonctdist   */
      classifiercode,          /* Stores code of classifier     */
      i;                       /* iterator variable             */
  LTUparmT
    *classifierparms=NULL;     /* Stores trained neural unit    */
  #ifdef VERBOSEPASIF
  int locusct=1;               /* Tracks count of input loci    */
  #endif
  #ifdef MPI
  int procrank;                /* Stores rank of processor      */

  MPI_Init(&argc,&argv);
  MPI_Comm_rank(MPI_COMM_WORLD,&procrank);
  #endif

  /* Verify that specified command line arguments are reasonable */
  if(argc<MINARGS) {
    #ifdef MPI
    if(procrank==MASTER)
    #endif
    BADUSE;
    exit(EXIT_FAILURE);
  }

  /* As setting the correct Markov model parameters is usually *
   * accident prone, it would be best to process this aspect   *
   * of the setup first.  First, allocate space for immprobT,  *
   * dmprobT, and foprobT instances, whether they will be used *
   * or not. (Note that you cannot just use a single void      *
   * pointer and set it to whatever type of parameter file     *
   * needed, as they can be used in varying combinations.)     */
  if((immprobs=(immprobT*)malloc(sizeof(immprobT)))==NULL ||
     (dmprobs=(dmprobT*)malloc(sizeof(dmprobT)))==NULL ||
     (fixordprobs=(foprobT*)malloc(sizeof(foprobT)))==NULL)
    fatalerror("Err (main): Insufficient Memory!\n");

  /* Determine what type of Markov model variant is *
   * intended for use, and read in the appropriate  *
   * parameterization data.                         */
  MMalgocode=atoi(argv[7]);
  switch(MMalgocode) {
    case FIXORDX :
    case TDDIX :
    case BUDIX :
    case CHI2X :
      if((test=strstr(argv[8],algotags[MMalgocode]))==NULL) {
        (void)snprintf(errbuff,MAXERRLEN,"Err (main): \
Binary parm file %s has improper extension!\n",argv[8]);
        fatalerror(errbuff);
      }
      else if(MMalgocode==FIXORDX)
        import_fo_probs(argv[8],fixordprobs);
      else /* an IMM variant: TDDI, BUDI, or CHI2 */
        import_imm_probs(argv[8],immprobs);
      break;
    case DMMMX :
      /* Quick check: if user is deploying the DMMM, determine if *
       * they want to use cloud bounded by range or variance.     */
      if(argc<(int)MINARGS+1) {
        (void)snprintf(errbuff,MAXERRLEN,DMUSAGE,argv[0],
          DMMMX,algotags[DMMMX],VARIANCEBASED,RANGEBASED);
        fatalerror(errbuff);
      }
      else
        cloudtype=atoi(argv[8]);

      if((test=strstr(argv[9],algotags[DMMMX]))==NULL) {
        (void)snprintf(errbuff,MAXERRLEN,"Err (main): \
Binary parm file %s has improper extension!\n",argv[9]);
        nonfatalerror(errbuff);
        (void)snprintf(errbuff,MAXERRLEN,DMUSAGE,argv[0],
          DMMMX,algotags[DMMMX],VARIANCEBASED,RANGEBASED);
        fatalerror(errbuff);
      }
      else
        import_dm_probs(argv[9],dmprobs);
      break;
    case PMMMX :
      if(argc<(int)MINARGS+2) {
        (void)snprintf(errbuff,MAXERRLEN,PMUSAGE,argv[0],PMMMX,
           algotags[DMMMX],algotags[FIXORDX],VARIANCEBASED,RANGEBASED);
        fatalerror(errbuff);
      }
      else
        cloudtype=atoi(argv[8]);

      if((test=strstr(argv[9],algotags[DMMMX]))==NULL ||
         (test=strstr(argv[10],algotags[FIXORDX]))==NULL) {
        (void)snprintf(errbuff,MAXERRLEN,"Err (main): \
Binary parm file (%s OR %s) has improper extension!\n",argv[9],argv[10]);
        nonfatalerror(errbuff);
        (void)snprintf(errbuff,MAXERRLEN,PMUSAGE,argv[0],PMMMX,
          algotags[DMMMX],algotags[FIXORDX],VARIANCEBASED,RANGEBASED);
        fatalerror(errbuff);
      }
      else {
        import_dm_probs(argv[9],dmprobs);
        import_fo_probs(argv[10],fixordprobs);
      }
      break;
    default :
      fatalerror("\nErr (main): \
Improper Markov model variant requested!\n\n");
  } /* end switch */

  /* Allocate space for Bssmparm instance, *
   * and attempt to read data into it.     */
  if((bssmparms=(Bssmparm*)malloc(sizeof(Bssmparm)))!=NULL)
    import_Bssmparm(argv[4],bssmparms);
  else
    fatalerror("Err (main): Insufficient Memory!\n");

  /* Determine which classification algorithm *
   * to use, and import its parameters.       */
  classifiercode=atoi(argv[5]);
  switch(classifiercode) {
    case LINEARUNITX :
    case SIGMOIDUNITX :
      if((test=strstr(argv[6],classifiertags[classifiercode]))==NULL) {
        (void)snprintf(errbuff,MAXERRLEN,"Err (main): \
LTU parm file %s has improper extension!\n",argv[6]);
        fatalerror(errbuff);
      }

      if((classifierparms=(LTUparmT*)malloc(sizeof(LTUparmT)))!=NULL)
        import_trained_perceptron(argv[6],classifierparms);
      else
        fatalerror("Err (main): Insufficient Memory!\n");
      break;
    case ORGOBJFCTX :
      /* There is no trained classifier for orgobjfct, so we simply  *
       * pass classifierparms to makeASforms as NULL. To keep comm   *
       * line option handling consistent, you are required to pass   *
       * the name of a "trained classifier file"---make up a name    *
       * if you like (since, again, it will be ignored), so long as  *
       * you pass "some" token for option counting to be consistent. */
      classifierparms=NULL;
      break;
    default :
      fatalerror("\nErr (main): \
Improper classifier method requested!\n\n");
  } /* end switch */

  /* Read in genomic loci for PASIF to process. */
  if((genfile=fopen(argv[1],"rt"))==NULL) {
    (void)snprintf(errbuff,MAXERRLEN,"Err (main): \
Can't seem to open %s!\n",argv[1]);
    fatalerror(errbuff);
  }

  /* Parse the exon count distribution and general config files */
  exonctdist=parseexonctdist(argv[2],exonctdist,&exonctdistsize);
  parsegeneralinfo(argv[3]); /* Modifies globals declared in pasif_utils.h */

  /* At this stage, all data that will be constant across all loci are in *
   * hand.  Any other necessary information will be learned in the while- *
   * loop below on a case-specific basis.                                 */

  /* Print XML header.  This can't be done in the makeASforms routine,  *
   * because it will be called multiple times to process multiple loci. */
  #ifdef MPI
  if(procrank==MASTER) {
  #endif
  fprintf(XMLOUTSTREAM,"<?xml \
version=\"1.0\" \
encoding=\"ISO-8859-1\"?>\n");
  fprintf(XMLOUTSTREAM,"<PASIF_output \
xmlns=\"http://www.pasif.org/PASIF_output/\" \
PASIF_XML_version=\"1.0\">\n");

  /* Lots of random number-dependent stuff occurs in *
   * functions called below, so we call srand now.   */
  srand((unsigned int)time(NULL));
  #ifdef MPI
  }
  #endif

  /* Process the input loci */
  while((charloc=get_fasta(genfile,charloc,&locuslength,locusid))!=NULL) {
    if(locuslength==0)
      if(charloc==NULL)
        fatalerror("Err (main): No real sequence passed!\n");
      else
        fatalerror("Err (main): Passed a 0nt genomic locus!\n");
    else if((locus=(int*)malloc(sizeof(int)*locuslength))==NULL)
      fatalerror("Err (main): Out of memory!\n");
    else { /* Copy charloc translation into locus */
      for(i=0;i<locuslength;++i)
        locus[i]=trans(charloc[i]);
      free(charloc);
    }

    /* Pre-processing step: Decorate the locus with a posteriori splice   *
     * site probabilities and rho-values computed under each of the donor *
     * and acceptor hypotheses.                                           */
    if((donscores=(double*)malloc(sizeof(double)*locuslength))==NULL ||
       (accscores=(double*)malloc(sizeof(double)*locuslength))==NULL ||
       (rho_vals_don=(double*)malloc(sizeof(double)*locuslength))==NULL ||
       (rho_vals_acc=(double*)malloc(sizeof(double)*locuslength))==NULL)
      fatalerror("Err (main): Out of memory!\n");
    else
      analyze_locus_splice_sites(bssmparms,locuslength,locus,donscores,
                                 accscores,rho_vals_don,rho_vals_acc);

    #ifdef VERBOSEPASIF
    /* For the impatient, log progress to stderr */
    (void)snprintf(errbuff,MAXERRLEN,
      "\n^^^ Processing locus %i ^^^\n",locusct++);
    nonfatalerror(errbuff);
    #endif

    /* Predict transcript isoforms. */
    makeASforms(locus,locuslength,locusid,exonctdist,exonctdistsize,
                donscores,accscores,rho_vals_don,rho_vals_acc,
                MMalgocode,fixordprobs,immprobs,dmprobs,cloudtype,
                classifiercode,classifierparms);

    /* Free memory */
    free(locus);
    free(donscores);
    free(accscores);
    free(rho_vals_don);
    free(rho_vals_acc);
  } /* end while */

  /* Finalize XML output */
  #ifdef MPI
  if(procrank==MASTER)
  #endif
  fprintf(XMLOUTSTREAM,"</PASIF_output>\n");

  #ifdef VERBOSEPASIF
  nonfatalerror("\n\nFinished with gene structure predictions!\n\n");
  #endif

  /* Close input stream and free memory */
  fclose(genfile);
  free(exonctdist);
  free(bssmparms);
  free(fixordprobs);
  free(immprobs);
  free(dmprobs);
  if(classifierparms!=NULL)
    free(classifierparms);

  #ifdef MPI
  MPI_Finalize();
  #endif

  return(EXIT_SUCCESS);
} /* end main */
