/* confparsing.c
 * Michael Sparks (mespar1@iastate.edu)
 * Last modified : 14 June 2007
 *
 * Contains functions used by PASIF for parsing configuration files
 *
 * Copyright (c) 2006,2007 Michael E Sparks
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <float.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "confparsing.h"
#include "errors.h"
#include "pasif_utils.h"

/* Function to parse exon count distribution file (filename), and store *
 * boundaries in an array of doubles (distbounds) for later use when    *
 * sampling from the unit interval.  filename should point to a char    *
 * array with the file to be parsed, and distbounds should be set to    *
 * NULL when this routine is called. Note the format of this config     *
 * file is as follows:                                                  *
 *   Lines beginning with '%' are treated as comments.                  *
 *   '%i\t%i' <- first token is number of exons in an alternatively     *
 *               spliced transcript (>=2, STRICTLY!), while second is   *
 *               the number of isoforms that exhibited this quality.    *
 *               Note that lines in this file MUST be in strictly       *
 *               increasing numeric order w.r.t. the first token.       *
 * size will record the largest index available to distbounds.          */
double *parseexonctdist(char *filename,double *distbounds,int *size) {
  FILE *conffile=NULL;       /* File pointer                      */
  int totalcts=0,            /* Number of isoforms reported       */
      highct=1,              /* Highest exon count seen so far    */
      exonct,                /* First token of record             */
      numwithexonct,         /* Second token of record            */
      *counts=NULL,          /* Counts of exons. (first two       *
                              * elements ignored!)                */
      i;                     /* iterator variable                 */
  char confline[MAXLINELEN]; /* Stores lines from conf file       */
  double *rfreqs=NULL;       /* Relative frequencies of different *
                              * with given exon ct. (first two    *
                              * elements ignored!)                */

  /* Open config file and initialize counts */
  if((conffile=fopen(filename,"rt"))==NULL) {
    (void)snprintf(errbuff,MAXERRLEN,"Err (parseexonctdist): \
Can't seem to open %s!\n",filename);
    fatalerror(errbuff);
  }
  else if((counts=(int*)malloc(sizeof(int)*2))==NULL)
    fatalerror("Err (parseexonctdist): \
Insufficient Memory!\n");

  /* Slurp the file */
  while(fgets(confline,MAXLINELEN,conffile)!=NULL) {
    if(confline[0]=='%') /* comment line */
      continue;
    else if(!sscanf(confline,"%i=%i\n",&exonct,&numwithexonct))
      fatalerror("Err (parseexonctdist): \
Error parsing formatted record!\n");
    else if(exonct<=highct)
      fatalerror("Err (parseexonctdist): \
Bad exon dist config file!\n");
    else if((counts=(int*)realloc(counts,sizeof(int)*(exonct+1)))==NULL)
      fatalerror("Err (parseexonctdist): Out of memory!\n");
    else {
      if(exonct-highct>1)
        for(i=highct+1;i<exonct;++i)
          counts[i]=0;
      counts[exonct]=numwithexonct;
      totalcts+=numwithexonct;
      highct=exonct;
    }
  }
  fclose(conffile);

  /* Populate rfreqs array */
  if((rfreqs=(double*)malloc(sizeof(double)*(highct+1)))==NULL)
    fatalerror("Err (parseexonctdist): Insufficient Memory!\n");
  else {
    rfreqs[0]=rfreqs[1]=0.0; /* These elts should never be used elsewhere! */
    for(i=2;i<=highct;++i)
      rfreqs[i]=rfreqs[i-1]+counts[i]/(double)totalcts;
    if(fabs(rfreqs[highct]-1.0)>DBL_EPSILON) /* sanity check */
      fatalerror("Err (parseexonctdist): \
Unexpected error occurred!\n");
  }

  /* Cleanup */
  free(counts);

  *size=highct;
  distbounds=rfreqs;
  return(distbounds);
} /* end parseexonctdist */

/* Function to parse general configuration information into *
 * the relevant data structures defined in pasif_utils.h.   */
void parsegeneralinfo(char *filename) {
  FILE *conffile=NULL;   /* File pointer             */
  char line[MAXLINELEN], /* Buffer for slurped lines */
       *chip;            /* used to process records  */

  /* Open config file */
  if((conffile=fopen(filename,"rt"))==NULL) {
    (void)snprintf(errbuff,MAXERRLEN,"Err (parsegeneralinfo): \
Can't seem to open %s!\n",filename);
    fatalerror(errbuff);
  }

  /* Read in the options. */
  while(fgets(line,MAXLINELEN,conffile)!=NULL) {
    if(line[0]==(int)'%')
      continue;
    else if((chip=strstr(line,lenlimittags[MININTRONLENX]))!=NULL) {
      chip=strrchr(line,(int)'='); ++chip;
      lengthlimits[MININTRONLENX]=atoi(chip);
    }
    else if((chip=strstr(line,lenlimittags[MAXINTRONLENX]))!=NULL) {
      chip=strrchr(line,(int)'='); ++chip;
      lengthlimits[MAXINTRONLENX]=atoi(chip);
    }
    else if((chip=strstr(line,lenlimittags[MINEXONLENX]))!=NULL) {
      chip=strrchr(line,(int)'='); ++chip;
      lengthlimits[MINEXONLENX]=atoi(chip);
    }
    else if((chip=strstr(line,lenlimittags[MAXEXONLENX]))!=NULL) {
      chip=strrchr(line,(int)'='); ++chip;
      lengthlimits[MAXEXONLENX]=atoi(chip);
    }
    else if((chip=strstr(line,RESTARTLIMITX))!=NULL) {
      chip=strrchr(line,(int)'='); ++chip;
      restartlimit=atoi(chip);
    }
    else if((chip=strstr(line,TARGETISOFORMCTX))!=NULL) {
      chip=strrchr(line,(int)'='); ++chip;
      targetisoformct=atoi(chip);
    }
    else if((chip=strstr(line,MAXEXONCTFULLEXPANDX))!=NULL) {
      chip=strrchr(line,(int)'='); ++chip;
      maxexonctfullexpand=atoi(chip);
    }
    else if((chip=strstr(line,MAXSOLNDEPTHX))!=NULL) {
      chip=strrchr(line,(int)'='); ++chip;
      maxsolndepth=atoi(chip);
    }
    else {
      (void)snprintf(errbuff,MAXERRLEN,"Err (parsegeneralinfo): \
Blatantly non-conformant crud in %s! : %s\n",filename,line);
      fatalerror(errbuff);
    }
  }

  /* Cleanup and return */
  fclose(conffile);
  return;
} /* end parsegeneralinfo */
