/* dm_utils.c
 * Michael E Sparks (mespar1@iastate.edu)
 * Last modified: 14 June 2007
 *
 * This file contains code for using dynamically modulating Markov models.
 *
 * Copyright (C) 2005,2006,2007 Michael E Sparks
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
#include <float.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errors.h"
#include "immpractical.h"
#include "simplex.h"

#define LPm             10 /* number of constraints in LP problem        */
#define LPn              4 /* number of nonbasic variables in LP problem */
#define LPsize (LPm+LPn+1) /* Basic LP data structure size unit          */

/* Function definitions ******************************************************/

/* Function to return log-likelihood of a test input sequence  *
 * under the model specified using the model argument.         *
 * If threeper is FALSE, the                                   *
 * function returns the non-periodic log-probability of the    *
 * test sequence.  If non-FALSE, this function handles those   *
 * cases where model is an element of {0,1,2} if shadow is     *
 * FALSE, or {0,1,2,3,4,5} if shadow is non-FALSE; it returns  *
 * the probability of the input sequence given its first base  *
 * occurs in the phase specified by the model argument.        *
 * (see immpractical.h for phase nomenclature).                */
long double DMMMprob(int *nt,int length,dmprobT *probs,
                     int cloudtype,int model,int threeper,
                     int shadow)
{
  int /* stores counts of oligos in input sequence */
      testcts[ALFSIZE][ALFSIZE][ALFSIZE]
             [ALFSIZE][ALFSIZE][ALFSIZE],
      N[LPsize],         /* Store nonbasic status of each variable */
      Ntmp[LPsize]={0,1,1,1,1,0,0,0,0,0,0,0,0,0,0}, /*  copy to N  */
      status,            /* record status from simplex             */
      relevantp,         /* is some history pertinent to test seq? */
      localmodel,        /* helps index three-periodic cases       */
      i,j,k,l,m,n,
      z,aa,bb;           /* iterator variables                     */
  double **A=NULL,       /* Matrix of nonbasic var coefficients    */
         Atmp[LPsize][LPsize] =
           {  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0,-1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0, 0,-1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0, 0, 0,-1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0, 0, 0, 0,-1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
              { 0,-1,-1,-1,-1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
           },             /* This matrix will be invariant (A=Atmp)  */
         b[LPsize],       /* constraints concerning basic vars       */
         c[LPsize],       /* Nonbasic variable coefficients that     *
                           * occur in objective function.            */
         v,               /* Obj fct's current value (not used)      */
         *solution=NULL;  /* Stores information on optimal solution  */
  long double logprob=0.0;/* Stores overall log-probability of input */

  /* Verify that model is given a reasonable value */
  if(shadow==FALSE) {
    if((threeper==FALSE&&(model<0||model>NONCODING)) ||
       (threeper==TRUE&&(model<0||model>2)&&model!=NONCODING)) {
      (void)snprintf(errbuff,MAXERRLEN,"Err (DMMMprob): \
Invalid model value (%i)\n",model);
      fatalerror(errbuff);
    }
  }
  else if(model<3||model>5) {
    (void)snprintf(errbuff,MAXERRLEN,"Err (DMMMprob): \
Invalid model value (%i)\n",model);
    fatalerror(errbuff);
  }

  /* Verify cloudtype is set reasonably */
  if(cloudtype!=VARIANCEBASED&&cloudtype!=RANGEBASED) {
    (void)snprintf(errbuff,MAXERRLEN,"Err (DMMMprob): \
Invalid cloudtype specified (%i)\n",cloudtype);
    fatalerror(errbuff);
  }

  /* Verify that we really have a sequence to test. If input is set to NULL, *
   * this conditional will short-circuit and strlen won't get called on it.  */
  if(nt==NULL)
    fatalerror("Err (DMMMprob): No real sequence passed!\n");
  else if(length==0)
    fatalerror("Err (DMMMprob): Passed a 0nt test sequence!\n");

  /* A is basically a 2d array--I have to develop it dynamically in *
   * order to keep my simplex implentation as generic as possible,  *
   * which is overall a very minor inconvenience, IMHO.             */
  if((A=(double**)malloc(sizeof(double*)*LPsize))!=NULL)
    for(i=0;i<LPsize;++i) {
      if((A[i]=(double*)malloc(sizeof(double)*LPsize))==NULL)
        fatalerror("Err (DMMMprob): Out of memory!\n");
    }
  else
    fatalerror("Err (DMMMprob): Out of memory!\n");

  /* The idea is to return the probability of the sequence given *
   * the first base occurs in the frame denoted by model. z will *
   * denote the frame the input is read from, and localmodel the *
   * set of transition probability estimates to use.             */
  localmodel=model;
  for(z=0;z<3;++z) {
    /* Initialize count (parameter) array */
    for(i=0;i<ALFSIZE;++i)
    for(j=0;j<ALFSIZE;++j)
    for(k=0;k<ALFSIZE;++k)
    for(l=0;l<ALFSIZE;++l)
    for(m=0;m<ALFSIZE;++m)
    for(n=0;n<ALFSIZE;++n)
      testcts[i][j][k][l][m][n]=0;

    /* Set parameters for this test instance */
    if(model==NONCODING||threeper==FALSE)
      for(i=0;i<length-5;++i)
        ++testcts[nt[i]][nt[i+1]][nt[i+2]][nt[i+3]][nt[i+4]][nt[i+5]];
    else
      for(i=z;i<length-5;i+=3)
        ++testcts[nt[i]][nt[i+1]][nt[i+2]][nt[i+3]][nt[i+4]][nt[i+5]];

    /* Find those oligomers that are relevant to this test instance */
    for(i=0;i<ALFSIZE;++i)
    for(j=0;j<ALFSIZE;++j)
    for(k=0;k<ALFSIZE;++k)
    for(l=0;l<ALFSIZE;++l)
    for(m=0;m<ALFSIZE;++m) {
      relevantp=0;
      for(n=0;n<ALFSIZE;++n)
        if(testcts[i][j][k][l][m][n]>0) {
          relevantp=1;
          break;
        }

      /* Executes if we've found a relevant oligomer */
      if(relevantp){
        /* Set everything up for simplex.  Note that variables x_1, *
         * x_2, x_3, and x_4 correspond to bases A,C,G,T (0,1,2,3), *
         * respectively.                                            */
        v=0;
        for(aa=0;aa<LPsize;++aa) {
          N[aa]=Ntmp[aa];
          for(bb=0;bb<LPsize;++bb)
            A[aa][bb]=Atmp[aa][bb];
          /* pad b and c with zeros prior to their "real" inits */
          b[aa]=0;
          c[aa]=0;
        }
        for(aa=1;aa<=4;++aa)
          c[aa]=testcts[i][j][k][l][m][aa-1];
        for(aa=5;aa<=11;aa+=2)
          if(cloudtype==RANGEBASED) {
            b[aa]=probs->max[localmodel][i][j][k][l][m][(aa-5)/2];
            b[aa+1]=-probs->min[localmodel][i][j][k][l][m][(aa-5)/2];
          }
          else { /* (cloudtype==VARIANCEBASED) */
            b[aa]=probs->muplus1sigma[localmodel][i][j][k][l][m][(aa-5)/2];
            b[aa+1]=-probs->muless1sigma[localmodel][i][j][k][l][m][(aa-5)/2];
          }
        b[13]=1;
        b[14]=-1;

        /* Solve using my simplex implementation */
        solution=NULL;
        solution=simplex(LPsize,1,N,A,b,c,&v,LPm,LPn,&status,solution);
        if(solution==NULL)
          fatalerror("Err (DMMMprob): \
Simplex unexpectedly failed to compute a solution!\n");

        /* Increment logprob appropriately */
        for(aa=0;aa<ALFSIZE;++aa)
          if(testcts[i][j][k][l][m][aa]>0)
            logprob+=testcts[i][j][k][l][m][aa]*log(solution[aa+1]);

        /* return to heap */
        free(solution);
      } /* end if relevantp */
    } /* end m for */

    /* update localmodel */
    if(model==NONCODING||threeper==FALSE)
      break;
    else if(shadow==FALSE)
      localmodel=((localmodel+1)%3);
    else /* (shadow==TRUE) */
      localmodel=((localmodel+1)%3)+3;
  } /* end localmodel for */

  /* back to the heap */
  for(i=0;i<LPsize;++i)
    free(A[i]);
  free(A);

  if(isinf(logprob))
    nonfatalerror("Warning (DMMMprob): \
Likelihood score grew to infinity!\n");

  return(logprob);
} /* end DMMMprob */

/* Function to import a trained dmprobT     *
 * object (a binary file) into main memory. */
void import_dm_probs(char *filename,dmprobT *probs)
{
  FILE *fptr=NULL; /* For connecting to binary input file */

  /* Open parameter file */
  if((fptr=fopen(filename,"rb"))==NULL) {
    (void)snprintf(errbuff,MAXERRLEN,"Err (import_dm_probs): \
Can't open %s for binary reading!\n",filename);
    fatalerror(errbuff);
  }
  /* Read in the dmprobT "object" */
  else if(fread(probs,sizeof(dmprobT),1,fptr)!=1) {
    (void)snprintf(errbuff,MAXERRLEN,"Err (import_dm_probs): \
Error reading from %s!\n",
      filename);
    fatalerror(errbuff);
  }
  else
    (void)fclose(fptr);

  return;
} /* end import_dm_probs */
