/* immpractical.c
 * Michael E Sparks (mespar1@iastate.edu)
 * Last modified: 14 June 2007
 *
 * This file contains the core code of the IMMpractical library.
 *
 * Copyright (C) 2005,2006,2007 Michael E Sparks
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errors.h"
#include "immpractical.h"

/* Function implementations ************************************************/

/* Function to return log-likelihood of a test input sequence, *
 * under the model specified using the model argument.         *
 * If threeper is FALSE, the                                   *
 * function returns the non-periodic log-probability of the    *
 * test sequence.  If non-FALSE, this function handles those   *
 * cases where model is an element of {0,1,2} if shadow is     *
 * FALSE, or {0,1,2,3,4,5} if shadow is non-FALSE; it returns  *
 * the probability of the input sequence given its first base  *
 * occurs in the phase specified by the model argument.        *
 * (see immpractical.h for phase nomenclature).                */
long double IMMprob(int *nt,int length,immprobT *probs,
                    int model,int threeper,int shadow)
{
  int i;               /* Iterator variable    */
  long double logprob; /* end result to return */

  /* Verify that model is given a reasonable value */
  if(shadow==FALSE) {
    if((threeper==FALSE&&(model<0||model>NONCODING)) ||
       (threeper==TRUE&&(model<0||model>2)&&model!=NONCODING)) {
      (void)snprintf(errbuff,MAXERRLEN,"Err (IMMprob): \
Invalid model value (%i)\n",model);
      fatalerror(errbuff);
    }
  }
  else if(model<3||model>5) {
    (void)snprintf(errbuff,MAXERRLEN,"Err (IMMprob): \
Invalid model value (%i)\n",model);
    fatalerror(errbuff);
  }

  /* Verify that we really have a sequence to test. If input is set to NULL, *
   * this conditional will short-circuit and strlen won't get called on it.  */
  if(nt==NULL)
    fatalerror("Err (IMMprob): No real sequence passed!\n");
  else if(length==0)
    fatalerror("Err (IMMprob): Passed a 0nt test sequence!\n");

  /* compute log-likelihood of the input string */
  for(i=0,logprob=0.0;i<MAXORDER;++i)
    switch(i) {
      case 0 : /* mononucleotide */
        logprob+=log(probs->prob1[model][nt[0]]);
        break;
      case 1 : /* dinucleotide */
        logprob+=log(probs->prob2[model][nt[0]]
                                        [nt[1]]);
        break;
      case 2 : /* trinucleotide */
        logprob+=log(probs->prob3[model][nt[0]]
                                        [nt[1]]
                                        [nt[2]]);
        break;
      case 3 : /* tetranucleotide */
        logprob+=log(probs->prob4[model][nt[0]]
                                        [nt[1]]
                                        [nt[2]]
                                        [nt[3]]);
        break;
      case 4 : /* pentanucleotide */
        logprob+=log(probs->prob5[model][nt[0]]
                                        [nt[1]]
                                        [nt[2]]
                                        [nt[3]]
                                        [nt[4]]);
        break;
      default :
        fatalerror("Err (IMMprob): Bad short case passed!\n");
    } /* end switch */

  /* This loop processes positions such that full *
   * MAXORDER length histories are available.     */
  if(threeper!=FALSE) /* Compute 3-periodic result */
    for(i=MAXORDER;i<length;++i)
      switch(MAXORDER) {
        case(0) :
          if(shadow==FALSE) /* Forward strand */
            logprob+=log(probs->prob1[(i-MAXORDER+model)%3]
              [nt[i]]);
          else /* Reverse strand */
            logprob+=log(probs->prob1[((i-MAXORDER+model)%3)+3]
              [nt[i]]);
          break;
        case(1) :
          if(shadow==FALSE) /* Forward strand */
            logprob+=log(probs->prob2[(i-MAXORDER+model)%3]
              [nt[i-1]][nt[i]]);
          else /* Reverse strand */
            logprob+=log(probs->prob2[((i-MAXORDER+model)%3)+3]
              [nt[i-1]][nt[i]]);
          break;
        case(2) :
          if(shadow==FALSE) /* Forward strand */
            logprob+=log(probs->prob3[(i-MAXORDER+model)%3]
              [nt[i-2]][nt[i-1]][nt[i]]);
          else /* Reverse strand */
            logprob+=log(probs->prob3[((i-MAXORDER+model)%3)+3]
              [nt[i-2]][nt[i-1]][nt[i]]);
          break;
        case(3) :
          if(shadow==FALSE) /* Forward strand */
            logprob+=log(probs->prob4[(i-MAXORDER+model)%3]
              [nt[i-3]][nt[i-2]][nt[i-1]][nt[i]]);
          else /* Reverse strand */
            logprob+=log(probs->prob4[((i-MAXORDER+model)%3)+3]
              [nt[i-3]][nt[i-2]][nt[i-1]][nt[i]]);
          break;
        case(4) :
          if(shadow==FALSE) /* Forward strand */
            logprob+=log(probs->prob5[(i-MAXORDER+model)%3]
              [nt[i-4]][nt[i-3]][nt[i-2]][nt[i-1]][nt[i]]);
          else /* Reverse strand */
            logprob+=log(probs->prob5[((i-MAXORDER+model)%3)+3]
              [nt[i-4]][nt[i-3]][nt[i-2]][nt[i-1]][nt[i]]);
          break;
        case(5) :
          if(shadow==FALSE) /* Forward strand */
            logprob+=log(probs->prob6[(i-MAXORDER+model)%3]
              [nt[i-5]][nt[i-4]][nt[i-3]][nt[i-2]][nt[i-1]][nt[i]]);
          else /* Reverse strand */
            logprob+=log(probs->prob6[((i-MAXORDER+model)%3)+3]
              [nt[i-5]][nt[i-4]][nt[i-3]][nt[i-2]][nt[i-1]][nt[i]]);
          break;
        default:
          fatalerror("Err (IMMprob): Invalid MAXORDER encountered!\n");
      } /* end switch */
  else /* Compute non-periodic result */
    for(i=MAXORDER;i<length;++i)
      switch(MAXORDER) {
        case(0) :
          logprob+=log(probs->prob1[model][nt[i]]);
          break;
        case(1) :
          logprob+=log(probs->prob2[model][nt[i-1]]
                                          [nt[i]]);
          break;
        case(2) :
          logprob+=log(probs->prob3[model][nt[i-2]]
                                          [nt[i-1]]
                                          [nt[i]]);
          break;
        case(3) :
          logprob+=log(probs->prob4[model][nt[i-3]]
                                          [nt[i-2]]
                                          [nt[i-1]]
                                          [nt[i]]);
          break;
        case(4) :
          logprob+=log(probs->prob5[model][nt[i-4]]
                                          [nt[i-3]]
                                          [nt[i-2]]
                                          [nt[i-1]]
                                          [nt[i]]);
          break;
        case(5) :
          logprob+=log(probs->prob6[model][nt[i-5]]
                                          [nt[i-4]]
                                          [nt[i-3]]
                                          [nt[i-2]]
                                          [nt[i-1]]
                                          [nt[i]]);
          break;
        default:
          fatalerror("Err (IMMprob): Invalid MAXORDER encountered!\n");
      } /* end switch */

  return(logprob);
} /* end IMMprob */

/* Function to import a trained immprobT    *
 * object (a binary file) into main memory. */
void import_imm_probs(char *filename,immprobT *probs)
{
  FILE *fptr=NULL; /* For connecting to binary input file */

  /* Open parameter file */
  if((fptr=fopen(filename,"rb"))==NULL) {
    (void)snprintf(errbuff,MAXERRLEN,"Err (import_imm_probs): \
Can't open %s for binary reading!\n",filename);
    fatalerror(errbuff);
  }
  /* Read in the immprobT "object" */
  else if(fread(probs,sizeof(immprobT),1,fptr)!=1) {
    (void)snprintf(errbuff,MAXERRLEN,"Err (import_imm_probs): \
Error reading from %s!\n",filename);
    fatalerror(errbuff);
  }
  else
    (void)fclose(fptr);

  return;
} /* end import_imm_probs */
