/* pasif_utils.c
 * Michael Sparks (mespar1@iastate.edu)
 * Last modified : 20 June 2007
 *
 * Contains functions used by PASIF for conducting (local heuristic)
 * searches through candidate gene structure space.
 *
 * Copyright (c) 2006,2007 Michael E Sparks
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "classifiers.h"
#include "errors.h"
#include "immpractical.h"
#include "pasif_utils.h"
#include "splice_utils.h"
#ifdef MPI
#include "mpi.h"
#endif

#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define MIN(a,b) ((a) < (b) ? (a) : (b))

/* Local (static) function prototypes---see definitions for documentation ****/

static int sampleexonct(double *exonctdist,int exonctdistsize);
static exoncoorsT *induceseed(
  exoncoorsT *isoform,int exonct,int seqlen,int *status);
static void quicksortint(int *list,int left,int right);
static int auxpartition(int *list,int left,int right);
static exoncoorsT* refinestructure(
  exoncoorsT *original,int exonct,data4makefeatvecT *dat4mfv,
  int classifiercode,LTUparmT *classifierparms,int *isoformgoodp,
  double *isoformscore);
static void scoreisoform(
  exoncoorsT *isoform,data4makefeatvecT *dat4mfv,
  int classifiercode,LTUparmT *classifierparms,
  int *isoformgoodp,double *isoformscore);
static featvecT makefeatvec(
  exoncoorsT *isoform,data4makefeatvecT *dat4mfv);
static double calczeta(
  double *donprobs,double *accprobs,int itrct,int itrofinterest);
static exoncoorsT** enumerateALLviablechildren(
  exoncoorsT **childlist,int *childct,
  exonterminusrangeT **expansionranges,
  int maxexprangeind,data4makefeatvecT *dat4mfv,
  exoncoorsT *currmaxhead,exoncoorsT *candidatehead,
  exoncoorsT *candidatecurr,int *exonstartlowbound,
  int *exonstoplowbound,int *bounds);
static exoncoorsT** enumerateSOMEviablechildren(
  exoncoorsT **childlist,int *childct,
  exonterminusrangeT **expansionranges,
  int maxexprangeind,data4makefeatvecT *dat4mfv,
  exoncoorsT *currmaxhead,exoncoorsT *candidatehead,
  exoncoorsT *candidatecurr,int *exonstartlowbound,
  int *exonstoplowbound,int *bounds);
#ifndef MPI
static int findcurrbestsol(
  exoncoorsT **childlist,int childct,
  data4makefeatvecT *dat4mfv,int classifiercode,
  LTUparmT *classifierparms,exoncoorsT *currmaxhead,
  int *isoformgoodp,double *currmaxscore);
#endif
static void findtranslationstartstop(
  exoncoorsT *currmaxhead,exoncoorsT *candidatehead,
  int transstartlowbound,int transstoplowbound,
  data4makefeatvecT *dat4mfv,int classifiercode,
  LTUparmT *classifierparms,int *isoformgoodp,
  double *currmaxscore);
static int solutionknownp(
  exoncoorsT *isoform,exoncoorsT **solutions,int structurect);
static void freeexpansionrange(exonterminusrangeT *range);
static void freeisoform(exoncoorsT *isoform);
static void dumppredictions(
  exoncoorsT **solutions,double *solscores,
  int isoformct,char *locusid,int printscorep);

/* Function definitions ******************************************************/

/* This is the main function responsible for coordinating the *
 * process of predicting alternative splice isoforms.         */
void makeASforms(
  int *locus,           /* int translation of locus      */
  int locuslength,      /* stores length of locus array  */
  char *locusid,        /* description of locus          */
  double *exonctdist,   /* distribution of exon counts   *
                         * in known transcript isoforms  */
  int exonctdistsize,   /* largest index of exonctdist   */
  /* Parameters for pre-computed splice site data        */
  double *donscores,    /* decoration of locus with      *
                         * posterior donor site probs    */
  double *accscores,    /* decoration of locus with      *
                         * posterior acceptor site probs */
  double *rho_vals_don, /* rho-values under donor id     */
  double *rho_vals_acc, /* rho-values under acceptor id  */
  /* Parameters for scoring induced coding sequence      */
  int MMalgotype,       /* Code for Markov model variant */
  foprobT *fixordprobs, /* FO parameters                 */
  immprobT *immprobs,   /* TDDI,BUDI,CHI2 parameters     */
  dmprobT *dmprobs,     /* DMMM parameters               */
  int cloudtype,        /* var- or range-based [DP]MMM?  */
  /* Parameters for instance classifier function         */
  int classifiercode,   /* Index of classifier algorithm */
  LTUparmT
    *classifierparms    /* Trained classifier object     */
) {
  int i=0,                    /* iterator variable                  */
      cyclect=0,              /* Tracks number of seed-and-refine   *
                               * cycles executed thus far.          */
      isoformct,              /* Tracks count of distinct isoforms  *
                               * generated at locus so far.         */
      exonct=0,               /* Number of exons in current isoform */
      isoformgoodp=FALSE,     /* Stores classification of instance  *
                               * (TRUE=positive, FALSE=negative)    */
      solutionknownp_res,     /* result of solutionknownp predicate */
      exonctsampledfor,       /* counter for sampling exon ct       */
      seedstatus=FALSE;       /* Flags whether we could induce      *
                               * exonct exons on locus              */
  data4makefeatvecT dat4mfv;  /* Stores info for making featvec     */
  double *isoformscores=NULL; /* Stores scores for isoforms         */
  exoncoorsT **isoforms=NULL, /* Stores alternatively spliced       *
                               * isoform gene structure information */
           *localstruct=NULL; /* Stores local structure that the    *
                               * search process will optimize.      */
  #ifdef MPI
  int procrank,               /* Stores rank of processor           */
      totprocs;               /* Stores total count of procs        */
  MPI_Status status;          /* For message reception status       */

  MPI_Comm_rank(MPI_COMM_WORLD,&procrank);
  MPI_Comm_size(MPI_COMM_WORLD,&totprocs);
  #endif

  /* Pack up a data4makefeatvecT object, so we don't *
   * have to pass around so many arguments in the    *
   * relevant function calls, set ugly global vars,  *
   * etc.                                            */
  dat4mfv.locus=locus;
  dat4mfv.locuslength=locuslength;
  dat4mfv.donscores=donscores;
  dat4mfv.accscores=accscores;
  dat4mfv.rho_vals_don=rho_vals_don;
  dat4mfv.rho_vals_acc=rho_vals_acc;
  dat4mfv.MMalgotype=MMalgotype;
  dat4mfv.fixordprobs=fixordprobs;
  dat4mfv.immprobs=immprobs;
  dat4mfv.dmprobs=dmprobs;
  dat4mfv.cloudtype=cloudtype;

  #ifdef MPI
  if(procrank==MASTER)
  #endif
  /* Allocate array of pointers to putative spliced isoforms. *
   * targetisoformct is set in the general config file.       */
  if(((isoforms=
        (exoncoorsT**)malloc(sizeof(exoncoorsT*)*targetisoformct))==NULL) ||
     ((isoformscores=
        (double*)malloc(sizeof(double)*targetisoformct))==NULL))
    fatalerror("Err (makeASforms): Insufficient Memory!\n");

  /* This loop executes seed-and-refine cycles, i.e., random restart   *
   * hillclimbing search.  cyclect prevents any thrashing in this loop *
   * to carry on indefinitely, and is set in the general conf file.    */
  for(cyclect=isoformct=0;cyclect<restartlimit;++cyclect) {

    #ifdef MPI
    if(procrank==MASTER) {
    #endif
    /* (Attempt to) induce a structural seed that will fit in the locus */
    exonctsampledfor=0;
    do {
      if(exonctsampledfor++>MAXTIMESTOSAMPLEEXONCT)
        break;
      exonct=sampleexonct(exonctdist,exonctdistsize);
    } while(locuslength<exonct*lengthlimits[MINEXONLENX]+
                        (exonct-1)*lengthlimits[MININTRONLENX]);

    seedstatus=FALSE;
    if(exonctsampledfor<=MAXTIMESTOSAMPLEEXONCT)
      localstruct=induceseed(localstruct,exonct,locuslength,&seedstatus);

    #ifdef MPI
    for(i=0;i<totprocs;++i)
      if(i!=MASTER)
        MPI_Send(&seedstatus,1,MPI_INT,i,MESGTAG,MPI_COMM_WORLD);
    } /* end master-only work */
    else
      MPI_Recv(&seedstatus,1,MPI_INT,MASTER,MESGTAG,MPI_COMM_WORLD,&status);

    if(procrank==MASTER) {
      if(seedstatus==TRUE) {
        for(i=0;i<totprocs;++i)
          if(i!=MASTER)
            MPI_Send(&exonct,1,MPI_INT,i,MESGTAG,MPI_COMM_WORLD);
      }
      else if(localstruct!=NULL) /* sanity check */
        fatalerror("Err (makeASforms): \
localstruct not NULL but seedstatus is FALSE?!!\n");
    } /* end master-only work */
    else
      if(seedstatus==TRUE)
        MPI_Recv(&exonct,1,MPI_INT,MASTER,MESGTAG,MPI_COMM_WORLD,&status);

    #ifdef MPIDBG
    /* This assumes that all processes have access to I/O facilities! *
     * This check is more for verifying that the parallel computer is *
     * working properly.                                              */
    fprintf(stderr,"Process %i, seedstatus %i, exonct %i, cyclect %i\n",
      procrank,seedstatus,exonct,cyclect);
    #endif

    /* All processes will be concerned with seedstatus' value,  *
     * since all will be needed in the refinestructure routine. */
    if(seedstatus==TRUE) {
    #else
    if(localstruct!=NULL) {
    #endif

      /* Do greedy, heuristic local search. */
      localstruct=refinestructure(localstruct,exonct,&dat4mfv,classifiercode,
        classifierparms,&isoformgoodp,&isoformscores[isoformct]);

      #ifdef MPI
      /* Only the master process will record solutions */
      if(procrank==MASTER) {
      #endif

      /* If the refined structure is deemed reportable by the embedded *
       * classifier, verify that it has not been previously recorded.  */
      if(isoformgoodp==TRUE &&
         (solutionknownp_res=solutionknownp(
             localstruct,isoforms,isoformct))==FALSE) {
        /* Record solution (its score will already be captured) */
        isoforms[isoformct++]=localstruct;

        #ifdef VERBOSEPASIF
        /* For the impatient, log progress to stderr */
        (void)snprintf(errbuff,MAXERRLEN,
          "  !!! Recorded isoform %i of (at most) %i !!!\n",
          isoformct,targetisoformct);
        nonfatalerror(errbuff);
        #endif
      }
      else
        /* Discard junk */
        freeisoform(localstruct);

      #ifdef MPI
      for(i=0;i<totprocs;++i)
        if(i!=MASTER)
          MPI_Send(&isoformct,1,MPI_INT,i,MESGTAG,MPI_COMM_WORLD);
      } /* end master-only work */
      else
        MPI_Recv(&isoformct,1,MPI_INT,MASTER,MESGTAG,MPI_COMM_WORLD,&status);
      #endif

      /* Reset the localstruct pointer for the next cycle: *
       * If an isoform was found in this cycle,            *
       * then it will have been recorded in isoforms.      *
       * Else, its memory will have been freed by now.     */
      localstruct=NULL;

      if(isoformct>=targetisoformct)
        break;
    } /* end if */
    #ifndef MPI
    else if(seedstatus!=FALSE) /* sanity check  */
      fatalerror("Err (makeASforms): \
localstruct NULL but seedstatus is true?!!\n");
    #endif
  } /* end cyclect for */

  /* Report findings. The final argument of dumppredictions can be *
   * set to FALSE so as not to print scores, TRUE, otherwise.      */
  #ifdef MPI
  if(procrank==MASTER) {
  #endif
  dumppredictions(isoforms,isoformscores,isoformct,locusid,TRUE);

  /* Clean up memory and return */
  for(i=0;i<isoformct;++i)
    freeisoform(isoforms[i]);
  free(isoforms);
  free(isoformscores);
  #ifdef MPI
  }
  #endif

  return;
} /* end makeASforms */

/* Return a (fixed) number of exons for an isoform structural seed */
static int sampleexonct(
  double *exonctdist, /* distribution of exon counts  *
                       * in known transcript isoforms */
  int exonctdistsize  /* largest index of exonctdist  */
) {
  int i; /* iterator variable */
  double randnum; /* Store randomly sampled number */
  const double ceiling = 1000.0; /* Upper limit on random number */

  /* Draw a random number from the unit interval: Calculate a random *
   * number in range [0,ceiling], then take percentage.              */
  randnum=(rand()%((int)ceiling+1))/ceiling;

  for(i=2;i<=exonctdistsize;++i)
    if(exonctdist[i]>randnum)
      break;
  i=(i>exonctdistsize?exonctdistsize:i); /* just to be safe! */

  return(i);
} /* end sampleexonct */

/* This function will randomly generate an initial "seed" gene  *
 * structure that will be refined by other functions in this    *
 * file.  The isoform argument will be set to point to a linked *
 * list of exonct exoncoorsT elements, and should be set to     *
 * NULL when this function is invoked.  Constraints on min/max  *
 * exon/intron length will be observed.                         */
static exoncoorsT *induceseed(exoncoorsT *isoform,int exonct,
                              int seqlen,int *status
) {
  exoncoorsT
    *currelt=NULL,   /* isoform will keep track of the head node of  *
                      * the linked list, while this variable will be *
                      * used to create/traverse it.                  */
    *prevelt=NULL;   /* also used to traverse linked list            */
  int i,j,           /* iterator variable                            */
      seedisbad,     /* flag variable on whether seed will germinate */
      *bounds=NULL,  /* Store exon bounds prior to assigning isoform */
      eltlength,     /* Temporarily store length of exon or intron   */
      seedshapect=0; /* Tracks how many attemtps have been made to   *
                      * induce the exonct exons on the locus.        */

  /* First, a discussion concerning strategy.                           *
   * Since the greedy property does not hold in the context of gene     *
   * structure prediction, random restart hill-climbing (the approach   *
   * taken by the PASIF algorithm) confers the ability to predict       *
   * heterogeneous gene structures at a given locus.  So, inducing      *
   * seeds is a critical aspect of exploring the GLOBAL solution space. */

  /* Randomly pick exon boundary positions */
  if((bounds=(int*)malloc(sizeof(int)*(exonct<<1)))==NULL)
    fatalerror("Err (induceseed): Insufficient Memory!\n");

  *status=TRUE;
  do {
    if(seedshapect>MAXTIMESTOSHAPESEED) {
      *status=FALSE;
      break;
    }

    seedisbad=FALSE;
    for(i=0;i<exonct<<1;++i)
      bounds[i]=rand()%seqlen;
    quicksortint(bounds,0,(exonct<<1)-1);

    /* Implementation note: as rand() does not return negative *
     * numbers, there's no need to check for them.             */

    /* Are all exons of allowed lengths? */
    for(i=0;i<exonct<<1;i+=2) {
      eltlength=bounds[i+1]-bounds[i]+1;
      if(eltlength>lengthlimits[MAXEXONLENX] ||
         eltlength<lengthlimits[MINEXONLENX]) {
        seedisbad=TRUE;
        break;
      }
    }
    /* Are all introns of allowed lengths? */
    if(seedisbad == FALSE)
      for(i=1;i<(exonct<<1)-1;i+=2) {
        eltlength=bounds[i+1]-bounds[i]-1;
        if(eltlength>lengthlimits[MAXINTRONLENX] ||
           eltlength<lengthlimits[MININTRONLENX]) {
          seedisbad=TRUE;
          break;
        }
      }

    ++seedshapect;
  } while(seedisbad==TRUE); /* confirm seed observes constraints */

  if(*status==FALSE) {
    free(bounds);
    return NULL;
  }

  /* Build up the linked list representing the seed */
  i=0;
  if((isoform=(exoncoorsT*)malloc(sizeof(exoncoorsT)))==NULL)
    fatalerror("Err (induceseed): Insufficient Memory!\n");
  prevelt=isoform;
  isoform->start=bounds[i++];
  isoform->stop=bounds[i++];
  for(j=1;j<exonct;++j) {
    if((currelt=(exoncoorsT*)malloc(sizeof(exoncoorsT)))==NULL)
      fatalerror("Err (induceseed): Insufficient Memory!\n");
    currelt->start=bounds[i++];
    currelt->stop=bounds[i++];
    prevelt->nextelt=currelt;
    prevelt=currelt;
  }
  currelt->nextelt=NULL;

  free(bounds);
  return(isoform);
} /* end induceseed */

/* quicksort for int arrays, standard issue CLRS2ed */
static void quicksortint(int *list,int left,int right)
{
  int pivot;

  if(left<right) {
    pivot=auxpartition(list,left,right);
    quicksortint(list,left,pivot-1);
    quicksortint(list,pivot+1,right);
  }
  else
    return;
} /* end quicksortint */

/* Auxilliary function for quicksortint */
static int auxpartition(int *list,int left,int right)
{
  int temp,       /* For swapping elements */
      compare2me, /* For comparing element *
                   * values (ct_pret).     */
      i,j;        /* iterator variables    */

  compare2me=list[right];
  i=left-1;

  for(j=left;j<=right-1;++j)
    if(list[j]<=compare2me) {
      temp=list[++i];
      list[i]=list[j];
      list[j]=temp;
    }

  temp=list[i+1];
  list[i+1]=list[right];
  list[right]=temp;

  return(i+1);
} /* end auxpartition */

/* This function will improve upon the structural seed passed to it   *
 * using a greedy heuristic local search. It will modify the original *
 * argument until a locally optimal solution, achieved through        *
 * application of permissible, local operators, is found, and will    *
 * return a pointer to this data structure.                           */
static exoncoorsT* refinestructure(
  exoncoorsT *original,       /* Seed to refine             */
  int exonct,                 /* Number of exons in this    *
                               * candidate form             */
  data4makefeatvecT *dat4mfv, /* Info for making featvec    */
  /* Parameters for instance classifier function            */
  int classifiercode,         /* Classifier evaluation fct  */
  LTUparmT *classifierparms,  /* Trained classifier object  */
  int *isoformgoodp,          /* classification of result   */
  double *isoformscore        /* score of refined structure */
) {
  double currmaxscore;      /* Score of best child node encountered    */
  exoncoorsT
    *currmax=original,      /* ptr to currently-best isoform structure */
    *currmaxhead=original,
    *candidate=NULL,        /* Can this child node out-score currmax?  */
    *candidatehead=NULL,
    *candidateprev=NULL,
    **viablechildren=NULL;  /* List of node expansion-based candidates */
  exonterminusrangeT
    **expansionranges=NULL, /* List of ranges node expansions cover    */
    *exrangeprev=NULL,      /* Help build exonterminusrangeT link list */
    *exrangecurr=NULL;      /* Help build exonterminusrangeT link list */
  int i,m,                  /* iterator variables                      */
      *boundsarray=NULL,    /* Used in asserting that exon/intron      *
                             * length constraints are satisfied.       */
      *lowboundsarray=NULL, /* Used in defining domain of splice sites *
                             * and translation start/stop sites.       */
      solndepth,            /* Track depth of hill-climbing search     */
      childct,              /* Viable child nodes at search depth      */
      childsamplect,        /* children to sample in heuristic setting */
      classclown,           /* So we don't clobber isoformgoodp        */
      bestischildp=FALSE;   /* Records whether child out-scored parent */
  #ifdef MPI
  int j,k,l,                /* iterator variables                      */
      procrank,             /* Stores rank of processor                */
      totprocs,             /* Stores total count of procs             */
      maxchildind,          /* Index of successor node                 */
      childduty,            /* Size of each proc's node responsibility */
      *childcoors=NULL;     /* Records coordinates of child nodes      */
  double *childscores=NULL, /* Records scores of child nodes           */
         maxchildscore;     /* Records score of currently best child   */
  exoncoorsT
    *childhelperptr=NULL,   /* Copy superior node into currmaxhead     */
    *parenthelperptr=NULL;  /* Copy superior node into currmaxhead     */
  MPI_Status status;        /* For message reception status            */

  MPI_Comm_rank(MPI_COMM_WORLD,&procrank);
  MPI_Comm_size(MPI_COMM_WORLD,&totprocs);

  if(procrank==MASTER) {
  #endif
  /* Build a linked list for the candidate, and *
   * copy currmax (the parental node) into it.  */
  if((candidate=(exoncoorsT*)malloc(sizeof(exoncoorsT)))==NULL)
    fatalerror("Err (refinestructure): Insufficient Memory!\n");
  else {
    candidate->start=currmax->start;
    candidate->stop=currmax->stop;
    candidatehead=candidateprev=candidate;
  }
  for(i=1;i<exonct;++i)
    if((candidate=(exoncoorsT*)malloc(sizeof(exoncoorsT)))==NULL)
      fatalerror("Err (refinestructure): Insufficient Memory!\n");
    else {
      currmax=currmax->nextelt;
      candidate->start=currmax->start;
      candidate->stop=currmax->stop;
      candidateprev->nextelt=candidate;
      candidateprev=candidate;
    }
  if(currmax->nextelt != NULL) /* sanity check */
    fatalerror("Err (refinestructure): Inconsistency Detected!\n");
  else
    candidate->nextelt=NULL;

  /* Reset pointers. Again, candidate will serve as a parental node *
   * to be expanded, while currmax will record the currently-best   *
   * solution.                                                      */
  candidate=candidatehead;
  currmax=currmaxhead;

  #ifdef MPI
  } /* end master-only work */
  /* All processes will need the candidate data structure allocated,  *
   * but only the master will have access to currmax, which is passed *
   * as NULL to every other process.                                  */
  else {
    if((candidate=(exoncoorsT*)malloc(sizeof(exoncoorsT)))==NULL) {
      /* This assumes that all processes have access to I/O facilities! */
      fprintf(stderr,"Err (refinestructure): \
Insufficient memory on node %i!\n",procrank);
      exit(EXIT_FAILURE);
    }
    else
      candidatehead=candidate;
    for(i=1;i<exonct;++i) {
      candidateprev=candidate;
      if((candidate=(exoncoorsT*)malloc(sizeof(exoncoorsT)))==NULL) {
        /* This assumes that all processes have access to I/O facilities! */
        fprintf(stderr,"Err (refinestructure): \
Insufficient memory on node %i!\n",procrank);
        exit(EXIT_FAILURE);
      }
      candidateprev->nextelt=candidate;
      candidateprev=candidate;
    }
    candidate->nextelt=NULL;
    candidate=candidatehead;
  } /* end else for slave processes */

  if(procrank==MASTER) {
  #endif
  /* score this seed structure (currmaxscore needs to be set) */
  scoreisoform(currmax,dat4mfv,classifiercode,
    classifierparms,&classclown,&currmaxscore);

  /* Used by the enumerate*viablechildren routines for constraint checking */
  if((boundsarray=(int*)malloc(sizeof(int)*(exonct<<1)))==NULL)
    fatalerror("Err (refinestructure): Insufficient Memory!\n");

  /* The first and last elements of this array are LOWER bounds on *
   * translation start and stop, respectively.  The remaining      *
   * elements are alternating LOWER bounds for donor and acceptor  *
   * sites of the structure (these relative to the current node).  */
  if((lowboundsarray=(int*)malloc(sizeof(int)*(exonct<<1)))==NULL)
    fatalerror("Err (refinestructure): Insufficient Memory!\n");

  /* Create an array for lists that will record the range over *
   * the solution space which the sequential node expansions   *
   * cover.  This will be useful for pruning previously        *
   * considered child nodes during the search.                 */
  if((expansionranges=(exonterminusrangeT**)malloc(
    sizeof(exonterminusrangeT*)*(int)maxsolndepth))==NULL)
    fatalerror("Err (refinestructure): Insufficient Memory!\n");

  #ifdef VERBOSEPASIF
  (void)snprintf(errbuff,MAXERRLEN,
    "\nSeeding hill-climbing search with structure of %i exons\n",exonct);
  nonfatalerror(errbuff);
  if(exonct<=(int)maxexonctfullexpand)
    nonfatalerror("  (exhaustive node expansions permitted)\n");
  else
    nonfatalerror("  (random sampling of child nodes required)\n");
  #endif

  #ifdef MPI
  } /* end master-only work */
  #endif

  /* This loop implements depth-limited search.     *
   * On each iteration, currmax is the parent node. */
  for(solndepth=0;solndepth<(int)maxsolndepth;++solndepth) {
    #ifdef MPI
    if(procrank==MASTER) { /* The master process enumerates viable children */
    #endif
    /* Configure lowboundsarray for currmax. We safeguard against indexing *
     * elements under zero, but assume we won't access elements above the  *
     * last in the locus array.  This is a reasonable assumption since the *
     * translation stop site of the seed structure is guaranteed to occur  *
     * prior to the 3' terminus of the locus (enforced in the induceseed   *
     * routine), and in the findtranslationstartstop method, we take care  *
     * not to overstep this boundary.                                      *
     * In parallel, we populate a list of ranges covered by node           *
     * expansions, with each element corresponding to an exon terminus.    */
    i=0;
    currmax=currmaxhead;
    /* Range of considered translation start sites */
    if((expansionranges[solndepth]=(exonterminusrangeT*)
         malloc(sizeof(exonterminusrangeT)))==NULL)
      fatalerror("Err (refinestructure): Insufficient Memory!\n");
    else {
      exrangeprev=exrangecurr=expansionranges[solndepth];
      lowboundsarray[i++]=exrangecurr->upstreamlim=
        MAX(currmax->start-TRANSUPSTREAMLIM,0); 
      exrangecurr->downstreamlim=currmax->start+TRANSDOWNSTREAMLIM;
    }
    /* Range near first donor (recall, we're dealing with exons, *
     * but this is equivalent to explicitly handling introns).   */
    if((exrangecurr=(exonterminusrangeT*)
         malloc(sizeof(exonterminusrangeT)))==NULL)
      fatalerror("Err (refinestructure): Insufficient Memory!\n");
    else {
      lowboundsarray[i++]=exrangecurr->upstreamlim=
        MAX(currmax->stop-SPLICEUPSTREAMLIM,0); 
      exrangecurr->downstreamlim=currmax->stop+SPLICEDOWNSTREAMLIM;
      exrangeprev->nextelt=exrangecurr;
      exrangeprev=exrangecurr;
    }
    while(currmax->nextelt!=NULL) {
      currmax=currmax->nextelt;
      for(m=0;m<2;++m) {
        if((exrangecurr=(exonterminusrangeT*)
             malloc(sizeof(exonterminusrangeT)))==NULL)
          fatalerror("Err (refinestructure): Insufficient Memory!\n");

        if(m==0) { /* Range near (upstream) acceptor */
          lowboundsarray[i++]=exrangecurr->upstreamlim=
            MAX(currmax->start-SPLICEUPSTREAMLIM,0); 
          exrangecurr->downstreamlim=currmax->start+SPLICEDOWNSTREAMLIM;
        }
        else { /* Range near (downstream) donor */
          lowboundsarray[i++]=exrangecurr->upstreamlim=
            MAX(currmax->stop-SPLICEUPSTREAMLIM,0); 
          exrangecurr->downstreamlim=currmax->stop+SPLICEDOWNSTREAMLIM;
        }
        exrangeprev->nextelt=exrangecurr;
        exrangeprev=exrangecurr;
      }
    }
    /* Revise last exon bound in terms of translation stop sites */
    lowboundsarray[--i]=exrangecurr->upstreamlim=
      MAX(currmax->stop-TRANSUPSTREAMLIM,0);
    exrangecurr->downstreamlim=currmax->stop+TRANSDOWNSTREAMLIM;
    exrangecurr->nextelt=NULL;

    /* Find a successor node:                                         *
     * This process involves compiling a list of child nodes to score *
     * and sort, then finding an appropriate successor node (here, on *
     * the basis of splicing configuration).                          *

     * A quick time out: Computational complexity is a MAJOR issue here!
     * Consider the following:

"""
CL-USER> (defun enum-powers (base max-expt)
	   ;; Enumerate powers from base**1 through base**max-expt
	   (defun enum-powers-aux (curr-expt)
	     (if (> curr-expt max-expt)
		 t
		 (progn
		   (format t "~d & ~:d \\\\~%"
			   curr-expt (expt base curr-expt))
		   (enum-powers-aux (1+ curr-expt)))))
	   (enum-powers-aux 1))
ENUM-POWERS
CL-USER> (enum-powers 11 15)
1 & 11 \\
2 & 121 \\
3 & 1,331 \\
4 & 14,641 \\
5 & 161,051 \\
6 & 1,771,561 \\
7 & 19,487,171 \\
8 & 214,358,881 \\
9 & 2,357,947,691 \\
10 & 25,937,424,601 \\
11 & 285,311,670,611 \\
12 & 3,138,428,376,721 \\
13 & 34,522,712,143,931 \\
14 & 379,749,833,583,241 \\
15 & 4,177,248,169,415,651 \\
T
CL-USER> 
"""

    * Processing of all child nodes at a given depth would seem feasible    *
    * for (being optimistic) at most 5 exons (assuming splice up/downstream *
    * limits of 5; note that there are 2*(exonct-1) many splice sites in    *
    * such a structure, so we must process, essentially, 11**8 nodes at     *
    * every depth level in the search tree).  Parallelization might permit  *
    * exhaustive enumeration of children for exonct's of 6-7 (being very    *
    * optimistic!), but it is clear that some heuristic strategy is needed, *
    * as exon counts of 15-20 or more are typical in certain eukaryotes.    *

    * The suboptimal strategy that we will have to live with is that we can *
    * only consider ALL child nodes of an expansion for exonct values of at *
    * most maxexonctfullexpand; anything larger and we will only sample AT  *
    * MOST 11**(2*(maxexonctfullexpand-1)) distinct child nodes from each   *
    * parent, pruning the search tree (on the basis of prior expansion      *
    * ranges) as in the exhaustive situation.  (At most, b/c we only try to *
    * obtain a distinct, viable child node 11**(2*(maxexonctfullexpand-1))  *
    * times, but this doesn't guarantee that every time we reach in the bag *
    * our efforts will be fruitful.)                                        */

    childct=0;
    if(exonct<=(int)maxexonctfullexpand)
      /* enumerateALLviablechildren is called with solndepth-1 b/c we don't *
       * want to prune back child nodes falling under the range of the      *
       * CURRENT expansion!  Passing both currmaxhead and candidatehead is  *
       * NOT superfluous for this (recursive) function.                     */
      viablechildren=enumerateALLviablechildren(
        viablechildren,&childct,expansionranges,solndepth-1,
        dat4mfv,currmaxhead,candidatehead,candidatehead,
        lowboundsarray,&lowboundsarray[1],boundsarray);
    else /* Find the best child we can using heuristic sampling */
      for(i=0,childsamplect=
                (int)pow((double)(SPLICEUPSTREAMLIM+SPLICEDOWNSTREAMLIM+1),
                         (double)(((int)maxexonctfullexpand-1)<<1));
          i<childsamplect;++i)
        viablechildren=enumerateSOMEviablechildren(
          viablechildren,&childct,expansionranges,solndepth-1,
          dat4mfv,currmaxhead,candidatehead,candidatehead,
          lowboundsarray,&lowboundsarray[1],boundsarray);

    #ifdef VERBOSEPASIF
    (void)snprintf(errbuff,MAXERRLEN,
      "  Depth %3i: Scoring %6i child nodes...",solndepth,childct);
    nonfatalerror(errbuff);
    #endif

    #ifdef MPI
    for(i=0;i<totprocs;++i)
      if(i!=MASTER)
        MPI_Send(&childct,1,MPI_INT,i,MESGTAG,MPI_COMM_WORLD);
    } /* end master-only work */
    else
      MPI_Recv(&childct,1,MPI_INT,MASTER,MESGTAG,MPI_COMM_WORLD,&status);
    #endif

    /* It is certainly possible to exhaust locally attainable solutions. */
    if(childct==0) {
      ++solndepth; /* needed for cleanup of expansionranges */
      break;
    }

    #ifdef MPI
    /* We have a list of childct many children.  Each process in  *
     * the system, MASTER node included, will be responsible for  *
     * scoring a chunk of these, and will report its results to   *
     * the MASTER.  If the workload to farm out is less than      *
     * MINNODESTOFARM, the MASTER will just score the full        *
     * generation of children.  The MASTER process is responsible *
     * for finding the best child and recording the solution.     */

    if(procrank==MASTER &&
       (childscores=(double*)malloc(sizeof(double)*childct))==NULL)
      fatalerror("Err (refinestructure): \
Insufficient Memory on MASTER node\n");

    if((childduty=childct/totprocs)>=(int)MINNODESTOFARM) {
      /* Allocate local memory on each process for its lot of children */
      if((childcoors=(int*)malloc(sizeof(int)*childduty*(exonct<<1)))==NULL ||
         (procrank!=MASTER &&
          (childscores=(double*)malloc(sizeof(double)*childduty))==NULL)) {
        /* This assumes that all processes have access to I/O facilities! */
        fprintf(stderr,"Err (refinestructure): \
Insufficient Memory on node %i!\n",procrank);
        exit(EXIT_FAILURE);
      }

      /* The master process will Send each process its own chunk of data. */
      for(i=0;i<totprocs;++i) {
        if(i==MASTER) /* Master process will do its own work later */
          continue;

        if(procrank==MASTER) {
          /* Populate childcoors for a slave process.  Coordinates  *
           * are copied into an array of ints (rather than a linked *
           * list of exoncoorsT elements) in order to simplify      *
           * message passing.                                       */
          j=0; /* This variable indexes elts of childcoors */
          for(k=i*childduty;k<(i+1)*childduty;++k) {
            candidate=viablechildren[k];
            for(l=0;l<exonct;++l) {
              childcoors[j++]=candidate->start;
              childcoors[j++]=candidate->stop;
              candidate=candidate->nextelt;
            }
            if(candidate!=NULL) { /* sanity check */
              (void)snprintf(errbuff,MAXERRLEN,"Err (refinestructure): \
Inconsistency detected on process %i!\n",MASTER);
              fatalerror(errbuff);
            }
          }
          MPI_Send(&childcoors[0],childduty*(exonct<<1),
            MPI_INT,i,MESGTAG,MPI_COMM_WORLD);
        }

        if(i==procrank) {
          /* Recv what the master process just assigned */
          MPI_Recv(&childcoors[0],childduty*(exonct<<1),
            MPI_INT,MASTER,MESGTAG,MPI_COMM_WORLD,&status);

          /* Decode and score each assigned child */
          j=0;
          for(k=0;k<childduty;++k) {
            candidate=candidatehead;
            for(l=0;l<exonct;++l) {
              candidate->start=childcoors[j++];
              candidate->stop=childcoors[j++];
              candidate=candidate->nextelt;
            }
            if(candidate!=NULL) { /* sanity check */
              /* This assumes that all processes *
               * have access to I/O facilities!  */
              (void)snprintf(errbuff,MAXERRLEN,"Err (refinestructure): \
Inconsistency detected on process %i!\n",procrank);
              fatalerror(errbuff);
            }
            scoreisoform(candidatehead,dat4mfv,classifiercode,
              classifierparms,&classclown,&childscores[k]);
          }
          /* Return scores to master */
          MPI_Send(&childscores[0],childduty,
            MPI_DOUBLE,MASTER,MESGTAG,MPI_COMM_WORLD);
        }
      } /* end for */
    } /* end if */

    /* The master process finds the successor, if it exists. */
    if(procrank==MASTER) {
      if(childduty>=(int)MINNODESTOFARM) {
        /* First, the master scores its own assignment... */
        for(i=MASTER*childduty;i<(MASTER+1)*childduty;++i)
          scoreisoform(viablechildren[i],dat4mfv,classifiercode,
            classifierparms,&classclown,&childscores[i]);
        /* ...and any unassigned nodes, if necessary. */
        for(i=totprocs*childduty;i<childct;++i)
          scoreisoform(viablechildren[i],dat4mfv,classifiercode,
            classifierparms,&classclown,&childscores[i]);

        /* Master receives all results */
        for(i=0;i<totprocs;++i) {
          if(i==MASTER)
            continue;
          else
            MPI_Recv(&childscores[i*childduty],childduty,
              MPI_DOUBLE,i,MESGTAG,MPI_COMM_WORLD,&status);

          #ifdef MPIDBG
          (void)snprintf(errbuff,MAXERRLEN,
            "    Successful Recv of %i scores from proc %i\n",
            childduty,i);
          nonfatalerror(errbuff);
          #endif
        } /* end for */
      } /* end if */
      else /* MASTER scores all nodes */
        for(i=0;i<childct;++i)
          scoreisoform(viablechildren[i],dat4mfv,classifiercode,
            classifierparms,&classclown,&childscores[i]);

      /* Identify the (a) best child node */
      maxchildscore=childscores[0];
      maxchildind=0;
      for(i=1;i<childct;++i)
        if(!(childscores[i]<maxchildscore)) {
          maxchildind=i;
          maxchildscore=childscores[i];
        }
      /* Was the best child at least as good as its parent? *
       * If so, it will become a parent.                    */
      if(!(maxchildscore<currmaxscore)) {
        bestischildp=TRUE;
        /* Copy the best child node to currmaxhead */
        parenthelperptr=currmaxhead;
        childhelperptr=viablechildren[maxchildind];
        while(parenthelperptr->nextelt!=NULL) {
          parenthelperptr->start=childhelperptr->start;
          parenthelperptr->stop=childhelperptr->stop;
          parenthelperptr=parenthelperptr->nextelt;
          childhelperptr=childhelperptr->nextelt;
        }
        parenthelperptr->start=childhelperptr->start;
        parenthelperptr->stop=childhelperptr->stop;
        if(childhelperptr->nextelt!=NULL) /* sanity check */
          fatalerror("Err (refinestructure): \
Inconsistency Detected!\n");
        /* Re-assign currmaxscore.  Don't worry about classclown (the *
         * classification status), as it will be taken care of in the *
         * findtranslationstartstop function after the solndepth for- *
         * loop.                                                      */
        currmaxscore=maxchildscore;
      }
      else
        bestischildp=FALSE;

      /* Inform all other processes of this state */
      for(i=0;i<totprocs;++i)
        if(i!=MASTER)
          MPI_Send(&bestischildp,1,MPI_INT,i,MESGTAG,MPI_COMM_WORLD);

      for(i=0;i<childct;++i)
        freeisoform(viablechildren[i]);
      free(viablechildren);
      viablechildren=NULL;
      if(childduty>=(int)MINNODESTOFARM)
        free(childcoors);
      free(childscores);
    } /* end master-only work */
    else {
      if(childduty>=(int)MINNODESTOFARM) {
        free(childcoors);
        free(childscores);
      }

      MPI_Recv(&bestischildp,1,MPI_INT,MASTER,MESGTAG,MPI_COMM_WORLD,&status);
    } /* end slave-only work */

    if(bestischildp!=TRUE) {
      ++solndepth; /* needed for cleanup of expansionranges *
                    * (though only by the master process)   */
      break;
    }
    #else
    /* At this point, we have memory bound in three distinct data  *
     * structures: viablechildren, currmaxhead, and candidatehead. *
     * In the following function, all of viablechildren will be    *
     * freed, and the best node (possibly a successor) will be     *
     * recorded in currmaxhead.                                    */
    if((bestischildp=findcurrbestsol(viablechildren,childct,
         dat4mfv,classifiercode,classifierparms,
         currmaxhead,&classclown,&currmaxscore))!=TRUE) {
      /* No child was as good or better than its parent */
      ++solndepth; /* needed for cleanup of expansionranges */
      break;
    }
    else
      viablechildren=NULL;
    #endif
    #ifdef VERBOSEPASIF
    (void)snprintf(errbuff,MAXERRLEN,
      "Best child score: %.4e\n",currmaxscore);
    nonfatalerror(errbuff);
    #endif
  } /* end depth-limiting for */
  #ifdef VERBOSEPASIF
  if(childct==0||bestischildp!=TRUE)
    nonfatalerror("No child can succeed its parent.\n");
  #endif

  #ifdef MPI
  if(procrank==MASTER) {
  #endif
  /* Now, determine locally optimal overall gene structure   *
   * boundaries, i.e., translation start and stop sites.     *
   * This is performed regardless of whether a successor was *
   * found on the basis of splicing configuration (which is  *
   * considered to be an independent subproblem).  Note that *
   * we already know currmaxscore from the previous step.    */
  findtranslationstartstop(currmaxhead,candidatehead,
    lowboundsarray[0],lowboundsarray[(exonct<<1)-1],dat4mfv,
    classifiercode,classifierparms,&classclown,&currmaxscore);
  /* These are guaranteed to be no worse than from  *
   * the parent node (see findtranslationstartstop) */
  *isoformgoodp=classclown;
  *isoformscore=currmaxscore;

  /* free memory and return refined structure */
  free(boundsarray);
  free(lowboundsarray);
  for(i=0;i<solndepth;++i)
    freeexpansionrange(expansionranges[i]);
  free(expansionranges);
  #ifdef MPI
  } /* end master-only work */
  #endif
  freeisoform(candidatehead);

  return(currmaxhead);
} /* end refinestructure */

/* Manages the task of scoring and classifying an isoform structure */
static void scoreisoform(
  exoncoorsT *isoform,
  data4makefeatvecT *dat4mfv,
  int classifiercode,
  LTUparmT *classifierparms,
  int *isoformgoodp,
  double *isoformscore
) {
  featvecT features=makefeatvec(isoform,dat4mfv);

  switch(classifiercode) {
    case LINEARUNITX :
    case SIGMOIDUNITX :
      *isoformscore=eval_activation_unit(
        features,*classifierparms,isoformgoodp,
        isoformscore,classifiercode);
      break;
    case ORGOBJFCTX :
      orgobjfct(features,isoformgoodp,isoformscore);
      break;
    default :
      fatalerror("\nErr (scoreisoform): \
Improper classifier method requested!\n\n");
  }

  return;
} /* end scoreisoform */

/* Computes a descriptive feature vector for the isoform argument */
static featvecT makefeatvec(
  exoncoorsT *isoform,       /* Candidate solution to score */
  data4makefeatvecT *dat4mfv /* Info for making featvec     */
) {
  double sumdonpost=0.0, /* sum of donor site posterior probs            */
         sumaccpost=0.0, /* sum of acceptor site posterior probs         */
         sumdonrho=0.0,  /* sum of donor site rho-value                  */
         sumaccrho=0.0,  /* sum of acceptor site rho-value               */
         *donprobs=NULL, /* Array of donor probs for the structure       */
         *accprobs=NULL, /* Array of acceptor probs for the structure    */
         squashable=0.0, /* used in computing zeta composite value       */
         priors[NUMMODELS],   /* Priors on various coding hypotheses     */
         loglikes[NUMMODELS], /* Store logged likelihood of codseq under *
                               * the NUMMODELS (here, 4) hypotheses.     */
         loglikerats[NUMMODELS], /* log-likelihood ratios w.r.t. the     *
                                  * NONCODING hypothesis                 */
         scaleupfac=0.0,         /* Scaling constant                     */
         constdenom=0.0,         /* Cache Pr(data)                       */
         cdsscore=0.0;           /* Overall posterior coding probability *
                                  * of the induced coding sequence       */
  exoncoorsT *solptr=NULL; /* use to traverse linked list                */
  featvecT locfeatvec;     /* Value to return                            */
  int *codseq=NULL,        /* Will store induced coding sequence         */
      cdslength=0,         /* record length of codseq (!= locuslength)   */
      exonct=1,            /* record number of exons in isoform          */
      i,j;                 /* iterator variables                         */

  /* Initialize this array to keep GCC happy. */
  for(i=0;i<NUMMODELS;++i)
    loglikes[i]=0;

  /* On the first pass over the linked list, we want to learn the CDS *
   * length, the number of exons, and during the process, to learn    *
   * sumdonpost, sumaccpost, sumdonrho, and sumaccrho. We will also   *
   * take this opportunity to build the donprobs and accprobs arrays, *
   * which are needed to compute zeta values.                         */
  cdslength=isoform->stop-isoform->start+1;
  sumdonpost=dat4mfv->donscores[isoform->stop+1];
  sumdonrho=dat4mfv->rho_vals_don[isoform->stop+1];
  if((donprobs=(double*)malloc(sizeof(double)))==NULL)
    fatalerror("Err (makefeatvec): Out of memory!\n");
  else
    donprobs[0]=dat4mfv->donscores[isoform->stop+1];
  solptr=isoform->nextelt;
  while(solptr->nextelt!=NULL) {
    cdslength+=solptr->stop-solptr->start+1;
    sumaccpost+=dat4mfv->accscores[solptr->start-1];
    sumaccrho+=dat4mfv->rho_vals_acc[solptr->start-1];
    sumdonpost+=dat4mfv->donscores[solptr->stop+1];
    sumdonrho+=dat4mfv->rho_vals_don[solptr->stop+1];
    if((donprobs=(double*)realloc(donprobs,sizeof(double)*++exonct))==NULL)
      fatalerror("Err (makefeatvec): Out of memory!\n");
    else
      donprobs[exonct-1]=dat4mfv->donscores[solptr->stop+1];
    if((accprobs=(double*)realloc(accprobs,sizeof(double)*(exonct-1)))==NULL)
      fatalerror("Err (makefeatvec): Out of memory!\n");
    else
      accprobs[exonct-2]=dat4mfv->accscores[solptr->start-1];
    solptr=solptr->nextelt;
  }
  cdslength+=solptr->stop-solptr->start+1;
  sumaccpost+=dat4mfv->accscores[solptr->start-1];
  sumaccrho+=dat4mfv->rho_vals_acc[solptr->start-1];
  if((accprobs=(double*)realloc(accprobs,sizeof(double)*exonct++))==NULL)
    fatalerror("Err (makefeatvec): Out of memory!\n");
  else
    accprobs[exonct-2]=dat4mfv->accscores[solptr->start-1];

  /* Populate average donor and acceptor probability, rho fields */
  locfeatvec.donmeanave=sumdonpost/(exonct-1);
  locfeatvec.accmeanave=sumaccpost/(exonct-1);
  locfeatvec.donmeanrho=sumdonrho/(exonct-1);
  locfeatvec.accmeanrho=sumaccrho/(exonct-1);

  /* Compute squashed sum of zeta-values for introns in isoform. *
   *                                                             *
   * Here is a special case to consider:                         *
   *                                                             *
   * Zeta for intron 0: 0.000000                                 *
   *   Don: 0.000001, Acc: 0.000001                              *
   * Zeta for intron 1: 0.000000                                 *
   *   Don: 0.000001, Acc: 0.000001                              *
   * Zeta for intron 2: 0.000000                                 *
   *   Don: 0.000001, Acc: 0.000001                              *
   *                                                             *
   * All of these sites score the default splice site prob,      *
   * 1E-6 (see DEFAULTSSPROB in splice_utils.h).  Thus, every    *
   * intron's zeta-value is 0.0, which is not particularly       *
   * informative.  The solution is, unless there exists an       *
   * intron in the structure with a splice site probability .GT. *
   * DEFAULTSSPROB, the composite value will be assigned         *
   * DEFAULTCOMPZETA (see pasif_utils.h).                        */
  locfeatvec.zetacomp=DEFAULTCOMPZETA;
  for(i=0;i<exonct-1;++i)
    if(donprobs[i]>DEFAULTSSPROB||accprobs[i]>DEFAULTSSPROB) {
      for(j=0,squashable=0.0;j<exonct-1;++j)
        squashable+=calczeta(donprobs,accprobs,exonct-1,j);
      locfeatvec.zetacomp=1.0/(1.0+exp(-1.0*squashable));
      break;
    }
  free(donprobs);
  free(accprobs);

  /* Compute posterior probability of the induced CDS */

  /* Copy CDS into codseq */
  if((codseq=(int*)malloc(sizeof(int)*cdslength))==NULL)
    fatalerror("Err (makefeatvec): Insufficient Memory!\n");

  for(solptr=isoform,j=0;solptr->nextelt!=NULL;solptr=solptr->nextelt)
    for(i=solptr->start;i<=solptr->stop;++i)
      codseq[j++]=dat4mfv->locus[i];
  for(i=solptr->start;i<=solptr->stop;++i)
    codseq[j++]=dat4mfv->locus[i];

  /* Calculate the logged posterior probability *
   * that codseq is coding, in any frame on the *
   * forward-sense strand.                      */
  switch(dat4mfv->MMalgotype) {
    case FIXORDX :
      for(i=0;i<NONCODING;++i)
        loglikes[i]=FOprob(codseq,cdslength,
                      dat4mfv->fixordprobs,i,TRUE,FALSE);
      loglikes[NONCODING]=
        FOprob(codseq,cdslength,
          dat4mfv->fixordprobs,NONCODING,FALSE,FALSE);
      break;
    case TDDIX :
    case BUDIX :
    case CHI2X :
      for(i=0;i<NONCODING;++i)
        loglikes[i]=IMMprob(codseq,cdslength,
                      dat4mfv->immprobs,i,TRUE,FALSE);
      loglikes[NONCODING]=
        IMMprob(codseq,cdslength,
          dat4mfv->immprobs,NONCODING,FALSE,FALSE);
      break;
    case DMMMX :
      for(i=0;i<NONCODING;++i)
        loglikes[i]=DMMMprob(codseq,cdslength,
                      dat4mfv->dmprobs,dat4mfv->cloudtype,i,TRUE,FALSE);
      loglikes[NONCODING]=
        DMMMprob(codseq,cdslength,
          dat4mfv->dmprobs,dat4mfv->cloudtype,NONCODING,FALSE,FALSE);
      break;
    case PMMMX :
      for(i=0;i<NONCODING;++i)
        loglikes[i]=DMMMprob(codseq,cdslength,
                      dat4mfv->dmprobs,dat4mfv->cloudtype,i,TRUE,FALSE);
      loglikes[NONCODING]=
        FOprob(codseq,cdslength,
          dat4mfv->fixordprobs,NONCODING,FALSE,FALSE);
      break;
    default :
      fatalerror("Err (makefeatvec): Improper algorithm requested!\n");
  } /* end switch */
  free(codseq);

  /* Specificy prior distribution on coding hypotheses */
  constdenom=(double)(((int)NUMMODELS-1)<<1);
  for(i=0;i<(int)NUMMODELS-1;++i)
    priors[i]=1.0/constdenom;
  priors[NONCODING]=0.5;

  /* Log-likelihood ratios help keep buffers in check... */
  for(i=0;i<NUMMODELS;++i)
    loglikerats[i]=loglikes[i]-loglikes[NONCODING];

  /* Scaling up the Bayes formula by a constant factor scaleupfac *
   * should *help* to prevent buffer under/overflow errors...     */
  for(i=0;i<NONCODING;++i)
    scaleupfac+=loglikerats[i]/NUMMODELS;

  for(i=0,constdenom=0.0;i<NUMMODELS;++i)
    constdenom+=priors[i]*exp(loglikerats[i]-scaleupfac);

  for(i=0,cdsscore=0.0;i<NONCODING;++i)
    cdsscore+=(priors[i]*exp(loglikerats[i]-scaleupfac))/constdenom;

  /* Record codprob field */
  locfeatvec.codprob=cdsscore;

  /* Feature vector is fully populated */
  return(locfeatvec);
} /* end makefeatvec */

/* Calculate the zeta-value for an intron of interest */
static double calczeta(
  double *donprobs, /* Donor site probabilities for introns    *
                     * of the host gene structure              */
  double *accprobs, /* Acceptor site probabilities for introns *
                     * of the host gene structure              */
  int itrct,        /* Number of introns in original gene      *
                     * structure, i.e., size of donprobs/      *
                     * accprobs                                */
  int itrofinterest /* Index of intron whose zeta-value is to  *
                     * be computed.                            */
) {
  double bestincluded=-1.0, /* Best sum of don/acc probs in set *
                             * with itrofinterest included.     */
         bestexcluded=-1.0, /* Best sum of don/acc probs in set *
                             * with itrofinterest excluded.     */
         score;             /* Score of elt of intron power set */
  int *inclusionp=NULL,     /* For elts of power set, record a  *
                             * 1 if intron's included, 0 o.w.   */
      elt,                  /* index of elt of power set        */
      eltSSct,              /* Per elt, tallies the number of   *
                             * splice sites it includes from    *
                             * from the original structure      */
      itr,                  /* Index for intron of some elt     */
      remainder,            /* used for binary decoding         */
      dividend;             /* used for binary decoding         */

  /* Allocate inclusionp buffer */
  if((inclusionp=(int*)malloc(sizeof(int)*itrct))==NULL)
    fatalerror("Err (calczeta): Insufficient Memory!\n");

  /* We would like to use itrct many nested for-loops to enumerate the  *
   * power set of introns, but unfortunately we can't say what itrct    *
   * is until run time!  Since the power set of elements is, by         *
   * definition, of cardinality 2**#elts, we can use just one for-loop  *
   * to implement the power set enumeration.  While it might seem       *
   * pointless to consider the case with elt==0 (perturbation with no   *
   * introns) this is meant in particular to handle the case when there *
   * is only one intron in the original structure.                      */
  for(elt=0;elt<1<<itrct;++elt) {

    /* elt denotes a unique entry in the power set; *
     * converting to binary will decode it.         */
    for(itr=0;itr<itrct;++itr) /* Blank the inclusionp array */
      inclusionp[itr]=0;
    for(dividend=elt;dividend!=0;dividend=dividend>>1) {
      remainder=dividend%2;
      inclusionp[--itr]=remainder;
    }
  
    /* The elt has been decoded at this point,  *
     * so score it as the mean average of donor *
     * and acceptor posterior probabilities of  *
     * those introns it includes.               */
    for(itr=eltSSct=0,score=0.0;itr<itrct;++itr) {
      score+=inclusionp[itr]*(donprobs[itr]+accprobs[itr]);
      eltSSct+=inclusionp[itr]<<1;
    }
    if(eltSSct>0)
      score/=eltSSct;

    /* classify appropriately. */
    if(inclusionp[itrofinterest]==0)
      bestexcluded=MAX(bestexcluded,score);
    else
      bestincluded=MAX(bestincluded,score);
  } /* end for */

  /* Free memory and return the zeta value */
  free(inclusionp);
  return((bestincluded-bestexcluded)/sqrt(itrct));
} /* end calczeta */

/* This recursive function enumerates all viable child nodes, *
 * given permissible local operators to apply to the parent   *
 * node (currmaxhead).  This enumeration is given in terms of *
 * splicing configuration only.                               */
static exoncoorsT** enumerateALLviablechildren(
  exoncoorsT **childlist,     /* List of viable child nodes          */
  int *childct,               /* Number of children in childlist     */
  exonterminusrangeT
    **expansionranges,        /* List of solution space coverages    */
  int maxexprangeind,         /* Hill-climbing search depth minus 1, *
                               * it is the last accessible index of  *
                               * the expansionranges data structure. */
  data4makefeatvecT *dat4mfv, /* For building feature vectors        */
  exoncoorsT *currmaxhead,    /* Head of currmax linked list         */
  exoncoorsT *candidatehead,  /* Head of candidate linked list       */
  exoncoorsT *candidatecurr,  /* Pointer to current elt of candidate */
  int *exonstartlowbound,     /* 5'-most possible exon start site    */
  int *exonstoplowbound,      /* 5'-most possible exon stop site     */
  int *bounds                 /* data structure for checking bounds  */
) {
  int locexonstartdomainind,  /* index over domain of donor site     */
      locexonstopdomainind,   /* index over domain of acceptor site  */
      candidateisbad,         /* flag variable                       */
      exonct,                 /* count num exons in full structure   */
      i,j,k;                  /* iterator variables                  */
  exoncoorsT
    *candidatehelperptr=NULL, /* For traversing candidate list       */
    *childptrprev=NULL,       /* used for building children          */
    *childptrcurr=NULL;       /* used for building children          */
  exonterminusrangeT
    *exrangehelper=NULL;      /* For traversing expansionranges elts */

  if(candidatecurr->nextelt!=NULL)
    if(candidatecurr==candidatehead) /* First elt of linked list: Keep *
                                      * translation start invariant    */
      /* For an upper bound on locexonstopdomainind, we do not invoke the *
       * MIN macro (against dat4mfv->locuslength), b/c this is checked    *
       * below.                                                           */
      for(locexonstopdomainind=exonstoplowbound[0];
          locexonstopdomainind <
            exonstoplowbound[0]+SPLICEUPSTREAMLIM+SPLICEDOWNSTREAMLIM+1;
          ++locexonstopdomainind) {
        candidatecurr->stop=locexonstopdomainind;

        /* Q: Why pass the second element of the [don|acc]lowbound arrays? *
         * A: See how enumerateALLviablechildren is called in the          *
         *    refinestructure function, and this should become clear.      */
        childlist=enumerateALLviablechildren(childlist,childct,
          expansionranges,maxexprangeind,dat4mfv,
          currmaxhead,candidatehead,candidatecurr->nextelt,
          &exonstartlowbound[2],&exonstoplowbound[2],bounds);
      }
    else
      for(locexonstartdomainind=exonstartlowbound[0];
          locexonstartdomainind <
            exonstartlowbound[0]+SPLICEUPSTREAMLIM+SPLICEDOWNSTREAMLIM+1;
          ++locexonstartdomainind) {
        candidatecurr->start=locexonstartdomainind;

        for(locexonstopdomainind=exonstoplowbound[0];
            locexonstopdomainind <
              exonstoplowbound[0]+SPLICEUPSTREAMLIM+SPLICEDOWNSTREAMLIM+1;
            ++locexonstopdomainind) {
          candidatecurr->stop=locexonstopdomainind;

          childlist=enumerateALLviablechildren(childlist,childct,
            expansionranges,maxexprangeind,dat4mfv,
            currmaxhead,candidatehead,candidatecurr->nextelt,
            &exonstartlowbound[2],&exonstoplowbound[2],bounds);
        }
      }
  else { /* Last elt of linked list: keep translation stop invariant */

    /* Copy the isoform's splice site and translation boundaries into *
     * the bounds array.  If constraints on exon/intron length limits *
     * are not met, or if trying to access elements of the locus      *
     * beyond upper bounds (lower bounds were guaranteed from         *
     * lowboundsarray), we prune this child node.                     */
    candidatehelperptr=candidatehead;
    exonct=1;
    i=0;
    while(candidatehelperptr->nextelt!=NULL) {
      bounds[i++]=candidatehelperptr->start;
      bounds[i++]=candidatehelperptr->stop;
      candidatehelperptr=candidatehelperptr->nextelt;
      ++exonct;
    }
    /* Note: candidatehelperptr==candidatecurr */
    bounds[i+1]=candidatehelperptr->stop; /* invariant, b/c we are *
                                           * optimizing w.r.t. the *
                                           * splicing config only  */

    /* Here, we set the bounds[i] element. */
    for(locexonstartdomainind=exonstartlowbound[0];
        locexonstartdomainind <
          exonstartlowbound[0]+SPLICEUPSTREAMLIM+SPLICEDOWNSTREAMLIM+1;
        ++locexonstartdomainind) {

      bounds[i]=candidatecurr->start=locexonstartdomainind;

      /* Verify that the node (isoform) hasn't been previously considered. *
       * Note that this will, by definition, exclude the parental node for *
       * all expansions but the first, which involves the seed structure.  */
      if(maxexprangeind==-1) /* The original seed is a parent here. */
        if(solutionknownp(candidatehead,&currmaxhead,1)) /* The seed */
          candidateisbad=TRUE;
        else
          candidateisbad=FALSE;
      else { /* The node's parent is distinct from the original seed. */
        candidateisbad=FALSE;
        for(j=0;j<=maxexprangeind;++j) {
          /* Prune the search tree of nodes encountered *
           * during previous expansions.                */
          exrangehelper=expansionranges[j];

          /* Note that each exon has two boundaries */
          for(k=0;k<exonct<<1;++k)
            if(bounds[k]<exrangehelper->upstreamlim ||
               bounds[k]>exrangehelper->downstreamlim)
              /* current range doesn't cover this instance */
              break;
            else
              exrangehelper=exrangehelper->nextelt;

          if(k>=exonct<<1) { /* .GT. never expected! */
            if(exrangehelper!=NULL) /* sanity check */
              fatalerror("Err (enumerateALLviablechildren): \
Inconsistency Detected!\n");
            else {
              candidateisbad=TRUE;
              break;
            }
          }
        } /* end for */
      } /* end else */
      if(candidateisbad==TRUE) /* Node already considered! :( */
        continue;

      /* Verify that constraints on exon/intron lengths are satisfied. *
       * Are all bounds in increasing genomic position ordering?       */
      for(j=0;j<(exonct<<1)-1;++j)
        if(bounds[j]>bounds[j+1]) {
          candidateisbad=TRUE;
          break;
        }
      /* Are all bounds under 3' limit of locus? */
      if(candidateisbad==FALSE)
        for(j=(exonct<<1)-1;j>=0;--j)
          if(bounds[j]>=dat4mfv->locuslength) {
            candidateisbad=TRUE;
            break;
          }
      /* Are all exons of allowed lengths? */
      if(candidateisbad==FALSE)
        for(j=0;j<exonct<<1;j+=2)
          if((bounds[j+1]-bounds[j]+1)>lengthlimits[MAXEXONLENX] ||
             (bounds[j+1]-bounds[j]+1)<lengthlimits[MINEXONLENX]) {
            candidateisbad=TRUE;
            break;
          }
      /* Are all introns of allowed lengths? */
      if(candidateisbad==FALSE)
        for(j=1;j<(exonct<<1)-1;j+=2)
          if((bounds[j+1]-bounds[j]-1)>lengthlimits[MAXINTRONLENX] ||
             (bounds[j+1]-bounds[j]-1)<lengthlimits[MININTRONLENX]) {
            candidateisbad=TRUE;
            break;
          }
      if(candidateisbad==TRUE) /* Constraints unsatisfied :( */
        continue;

      /* If we get this far, then the node is a viable successor  *
       * of its parent, so we push it on the list of child nodes: */
      if((childlist=(exoncoorsT**)realloc(childlist,
          sizeof(exoncoorsT*)*(*childct+1)))==NULL) {
        (void)snprintf(errbuff,MAXERRLEN,"Err (enumerateALLviablechildren): \
Insufficient Memory for %i children!\n",*childct);
        fatalerror(errbuff);
      }
      else
        *childct+=1;
      /* Now, copy in the child */
      if((childlist[*childct-1]=
           (exoncoorsT*)malloc(sizeof(exoncoorsT)))==NULL)
        fatalerror("Err (enumerateALLviablechildren): \
Insufficient Memory!\n");
      else {
        childptrprev=childptrcurr=childlist[*childct-1];
        candidatehelperptr=candidatehead;
        childptrcurr->start=candidatehelperptr->start;
        childptrcurr->stop=candidatehelperptr->stop;
      }
      for(j=1;j<exonct;++j)
        if((childptrcurr=(exoncoorsT*)malloc(sizeof(exoncoorsT)))==NULL)
          fatalerror("Err (enumerateALLviablechildren): \
Insufficient Memory!\n");
        else {
          candidatehelperptr=candidatehelperptr->nextelt;
          childptrcurr->start=candidatehelperptr->start;
          childptrcurr->stop=candidatehelperptr->stop;
          childptrprev->nextelt=childptrcurr;
          childptrprev=childptrcurr;
        }
      if(candidatehelperptr->nextelt!=NULL) /* sanity check */
        fatalerror("Err (enumerateALLviablechildren): \
Inconsistency Detected!\n");
      else
        childptrcurr->nextelt=NULL;
    } /* end locexonstartdomainind for */
  } /* end else */

  return(childlist);
} /* end enumerateALLviablechildren */

/* This recursive function randomly samples viable child nodes, *
 * without replacement, given permissible local operators to    *
 * apply to the parent node (currmaxhead).  Children are given  *
 * in terms of splicing configuration only.                     */
static exoncoorsT** enumerateSOMEviablechildren(
  exoncoorsT **childlist,     /* List of viable child nodes          */
  int *childct,               /* Number of children in childlist     */
  exonterminusrangeT
    **expansionranges,        /* List of solution space coverages    */
  int maxexprangeind,         /* Hill-climbing search depth minus 1, *
                               * it is the last accessible index of  *
                               * the expansionranges data structure. */
  data4makefeatvecT *dat4mfv, /* For building feature vectors        */
  exoncoorsT *currmaxhead,    /* Head of currmax linked list         */
  exoncoorsT *candidatehead,  /* Head of candidate linked list       */
  exoncoorsT *candidatecurr,  /* Pointer to current elt of candidate */
  int *exonstartlowbound,     /* 5'-most possible exon start site    */
  int *exonstoplowbound,      /* 5'-most possible exon stop site     */
  int *bounds                 /* data structure for checking bounds  */
) {
  int candidateisbad,         /* flag variable                       */
      exonct,                 /* count num exons in full structure   */
      i,j;                    /* iterator variables                  */
  exoncoorsT
    *candidatehelperptr=NULL, /* For traversing candidate list       */
    *childptrprev=NULL,       /* used for building children          */
    *childptrcurr=NULL;       /* used for building children          */
  exonterminusrangeT
    *exrangehelper=NULL;      /* For traversing expansionranges elts */

  if(candidatecurr->nextelt!=NULL)
    if(candidatecurr==candidatehead) { /* First elt of linked list: keep *
                                        * translation start invariant    */
      candidatecurr->stop=rand()%
        (SPLICEUPSTREAMLIM+SPLICEDOWNSTREAMLIM+1)+exonstoplowbound[0];

      /* Q: Why pass the second element of the [don|acc]lowbound arrays? *
       * A: See how enumerateSOMEviablechildren is called in the         *
       *    refinestructure function, and this should become clear.      */
      childlist=enumerateSOMEviablechildren(childlist,childct,
        expansionranges,maxexprangeind,dat4mfv,
        currmaxhead,candidatehead,candidatecurr->nextelt,
        &exonstartlowbound[2],&exonstoplowbound[2],bounds);
    }
    else {
      candidatecurr->start=rand()%
        (SPLICEUPSTREAMLIM+SPLICEDOWNSTREAMLIM+1)+exonstartlowbound[0];
      candidatecurr->stop=rand()%
        (SPLICEUPSTREAMLIM+SPLICEDOWNSTREAMLIM+1)+exonstoplowbound[0];

      childlist=enumerateSOMEviablechildren(childlist,childct,
        expansionranges,maxexprangeind,dat4mfv,
        currmaxhead,candidatehead,candidatecurr->nextelt,
        &exonstartlowbound[2],&exonstoplowbound[2],bounds);
    }
  else { /* Last elt of linked list: keep translation stop invariant */

    /* Copy the isoform's splice site and translation boundaries into *
     * the bounds array.  If constraints on exon/intron length limits *
     * are not met, or if trying to access elements of the locus      *
     * beyond upper bounds (lower bounds were guaranteed from         *
     * lowboundsarray), we prune this child node.                     */
    candidatehelperptr=candidatehead;
    exonct=1;
    i=0;
    while(candidatehelperptr->nextelt!=NULL) {
      bounds[i++]=candidatehelperptr->start;
      bounds[i++]=candidatehelperptr->stop;
      candidatehelperptr=candidatehelperptr->nextelt;
      ++exonct;
    }
    bounds[i++]=candidatecurr->start=rand()%
      (SPLICEUPSTREAMLIM+SPLICEDOWNSTREAMLIM+1)+exonstartlowbound[0];
    bounds[i]=candidatehelperptr->stop; /* invariant (translation stop) */

    /* We do not call solutionknownp to assert that the sampled child *
     * is unique.  The intuition behind this is that it is cheaper to *
     * score a few duplicated nodes than to check for redundancy of   *
     * each and every sample.                                         */

    /* Verify that the node (isoform) hasn't been previously considered. *
     * Note that this will, by definition, exclude the parental node for *
     * all expansions but the first, which involves the seed structure.  */
    if(maxexprangeind==-1) /* The original seed is a parent here. */
      if(solutionknownp(candidatehead,&currmaxhead,1)) /* The seed */
        candidateisbad=TRUE;
      else
        candidateisbad=FALSE;
    else { /* The node's parent is distinct from the original seed. */
      candidateisbad=FALSE;
      for(i=0;i<=maxexprangeind;++i) {
        /* Prune the search tree of nodes encountered *
         * during previous expansions.                */
        exrangehelper=expansionranges[i];
        /* Note that each exon has two boundaries */
        for(j=0;j<exonct<<1;++j)
          if(bounds[j]<exrangehelper->upstreamlim ||
             bounds[j]>exrangehelper->downstreamlim)
            /* current range doesn't cover this instance */
            break;
          else
            exrangehelper=exrangehelper->nextelt;

        if(j>=exonct<<1) { /* .GT. never expected! */
          if(exrangehelper!=NULL) /* sanity check */
            fatalerror("Err (enumerateSOMEviablechildren): \
Inconsistency Detected!\n");
          else {
            candidateisbad=TRUE;
            break;
          }
        }
      } /* end for */
    } /* end else */
    if(candidateisbad==TRUE) /* Node already considered! :( */
      return(childlist);

    /* Verify that constraints on exon/intron lengths are satisfied. *
     * Are all bounds in increasing genomic position ordering?       */
    for(i=0;i<(exonct<<1)-1;++i)
      if(bounds[i]>bounds[i+1]) {
        candidateisbad=TRUE;
        break;
      }
    /* Are all bounds under 3' limit of locus? */
    if(candidateisbad==FALSE)
      for(i=(exonct<<1)-1;i>=0;--i)
        if(bounds[i]>=dat4mfv->locuslength) {
          candidateisbad=TRUE;
          break;
        }
    /* Are all exons of allowed lengths? */
    if(candidateisbad==FALSE)
      for(i=0;i<exonct<<1;i+=2)
        if((bounds[i+1]-bounds[i]+1)>lengthlimits[MAXEXONLENX] ||
           (bounds[i+1]-bounds[i]+1)<lengthlimits[MINEXONLENX]) {
          candidateisbad=TRUE;
          break;
        }
    /* Are all introns of allowed lengths? */
    if(candidateisbad==FALSE)
      for(i=1;i<(exonct<<1)-1;i+=2)
        if((bounds[i+1]-bounds[i]-1)>lengthlimits[MAXINTRONLENX] ||
           (bounds[i+1]-bounds[i]-1)<lengthlimits[MININTRONLENX]) {
          candidateisbad=TRUE;
          break;
        }
    if(candidateisbad==TRUE) /* Constraints unsatisfied :( */
      return(childlist);

    /* If we get this far, then the node is a viable successor  *
     * of its parent, so we push it on the list of child nodes: */
    if((childlist=(exoncoorsT**)realloc(childlist,
        sizeof(exoncoorsT*)*(*childct+1)))==NULL) {
      (void)snprintf(errbuff,MAXERRLEN,"Err (enumerateSOMEviablechildren): \
Insufficient Memory for %i children!\n",*childct);
      fatalerror(errbuff);
    }
    else
      *childct+=1;
    /* Now, copy in the child */
    if((childlist[*childct-1]=
        (exoncoorsT*)malloc(sizeof(exoncoorsT)))==NULL)
      fatalerror("Err (enumerateSOMEviablechildren): \
Insufficient Memory!\n");
    else {
      childptrprev=childptrcurr=childlist[*childct-1];
      candidatehelperptr=candidatehead;
      childptrcurr->start=candidatehelperptr->start;
      childptrcurr->stop=candidatehelperptr->stop;
    }
    for(i=1;i<exonct;++i)
      if((childptrcurr=(exoncoorsT*)malloc(sizeof(exoncoorsT)))==NULL)
        fatalerror("Err (enumerateSOMEviablechildren): \
Insufficient Memory!\n");
      else {
        candidatehelperptr=candidatehelperptr->nextelt;
        childptrcurr->start=candidatehelperptr->start;
        childptrcurr->stop=candidatehelperptr->stop;
        childptrprev->nextelt=childptrcurr;
        childptrprev=childptrcurr;
      }
    if(candidatehelperptr->nextelt!=NULL) /* sanity check */
      fatalerror("Err (enumerateSOMEviablechildren): \
Inconsistency Detected!\n");
    else
      childptrcurr->nextelt=NULL;
  } /* end else */

  return(childlist);
} /* end enumerateSOMEviablechildren */

#ifndef MPI
/* This function will free all storage in childlist, while attempting *
 * to find a successor node scoring at least as well as its parent    *
 * (returns TRUE if so, FALSE o.w.).  At any rate, the currently-best *
 * solution will be recorded in currmaxhead, whose classification     *
 * status will be stored in isoformgoodp, and score in currmaxscore.  *
 * In finding a local optimum, if we can do only worse than a parent  *
 * node, we select the parent.  If on a plateau, we select a child    *
 * (whose configuration will be distinct from its parent) with the    *
 * hope that one of its descendants will score better than it did.    */
static int findcurrbestsol(
  exoncoorsT **childlist,     /* List of viable child nodes       */
  int childct,                /* Number of children in childlist  */
  data4makefeatvecT *dat4mfv, /* Info for making featvec          */
  int classifiercode,         /* Classifier evaluation fct        */
  LTUparmT *classifierparms,  /* Trained classifier object        */
  exoncoorsT *currmaxhead,    /* Pointer to currently-best result */
  int *isoformgoodp,          /* classification of best result    */
  double *currmaxscore        /* Score of best child seen so far  */
) {
  double
    candidatescore,        /* Score of child being considered     */
    bestchildscore;        /* Best score among the children       */
  exoncoorsT
    *childhelperptr=NULL,  /* Copy superior node into currmaxhead */
    *parenthelperptr=NULL; /* Copy superior node into currmaxhead */
  int
    childexceededp=FALSE,  /* flag whether child outscored parent */
    classclown,            /* So we don't clobber isoformgoodp    */
    bestchildind,          /* Stores index of the best child node */
    bestchildclass,        /* Stores class of the best child node */
    i;                     /* iterator variable                   */ 

  /* First, find the (a) best child node, and determine *
   * whether its score exceeds that of its parent.      */
  bestchildind=0;
  scoreisoform(childlist[0],dat4mfv,classifiercode,
    classifierparms,&bestchildclass,&bestchildscore);
  for(i=1;i<childct;++i) { 
    scoreisoform(childlist[i],dat4mfv,classifiercode,
      classifierparms,&classclown,&candidatescore);
    if(!(candidatescore<bestchildscore)) {
      bestchildind=i;
      bestchildclass=classclown;
      bestchildscore=candidatescore;
    }
  }
  if(!(bestchildscore<*currmaxscore))
    childexceededp=TRUE;

  if(childexceededp==TRUE) {
    /* Copy succeeding child as current maximum */
    parenthelperptr=currmaxhead;
    childhelperptr=childlist[bestchildind];
    while(parenthelperptr->nextelt!=NULL) {
      parenthelperptr->start=childhelperptr->start;
      parenthelperptr->stop=childhelperptr->stop;
      parenthelperptr=parenthelperptr->nextelt;
      childhelperptr=childhelperptr->nextelt;
    }
    parenthelperptr->start=childhelperptr->start;
    parenthelperptr->stop=childhelperptr->stop;
    if(childhelperptr->nextelt!=NULL) /* sanity check */
      fatalerror("Err (findcurrbestsol): Inconsistency Detected!\n");
    *currmaxscore=bestchildscore;
    *isoformgoodp=bestchildclass;
  }

  /* cleanup memory */
  for(i=0;i<childct;++i)
    freeisoform(childlist[i]);
  free(childlist);
  childlist=NULL;

  return(childexceededp);
} /* end findcurrbestsol */
#endif

/* This function finds the optimal descendant node on the  *
 * basis of translation start and stop site configuration. *
 * When called, the candidatehead structure should be set  *
 * to that of currmaxhead.                                 */
static void findtranslationstartstop(
  exoncoorsT *currmaxhead,    /* Head of currmax linked list           */
  exoncoorsT *candidatehead,  /* Head of candidate linked list         */
  int transstartlowbound,     /* 5'-most index of possible translation *
                               * start site                            */
  int transstoplowbound,      /* 5'-most index of possible translation *
                               * stop site                             */
  data4makefeatvecT *dat4mfv, /* Info for making featvec               */
  int classifiercode,         /* Classifier evaluation fct             */
  LTUparmT *classifierparms,  /* Trained classifier object             */
  int *isoformgoodp,          /* classification of result              */
  double *currmaxscore        /* Score of best child node seen so far  */
) {
  exoncoorsT
    *candidatetail=candidatehead, /* Set to tail of candidate list      */
    *currmaxtail=currmaxhead;     /* Set to tail of currmax list        */
  int loctransstartdomainind,     /* 5'-most index of translation start */
      loctransstopdomainind,      /* 5'-most index of translation stop  */
      classclown;                 /* So we don't clobber isoformgoodp   */
  double candidatescore;          /* Score of potential child node      */

  /* First, copy the currmaxhead structure into candidatehead, *
   * which will, coincidentally, initialize the tail pointers. *
   * This is a non-recursive function, so this is safe.        */
  while(currmaxtail->nextelt!=NULL) {
    candidatetail->start=currmaxtail->start;
    candidatetail->stop=currmaxtail->stop;
    currmaxtail=currmaxtail->nextelt;
    candidatetail=candidatetail->nextelt;
  }
  candidatetail->start=currmaxtail->start;
  candidatetail->stop=currmaxtail->stop;
  if(candidatetail->nextelt!=NULL) /* sanity check */
    fatalerror("Err (findtranslationstartstop): \
Inconsistency Detected!\n");

  for(loctransstartdomainind=transstartlowbound;
      loctransstartdomainind <
        MIN((transstartlowbound+TRANSUPSTREAMLIM+TRANSDOWNSTREAMLIM+1),
            dat4mfv->locuslength);
      loctransstartdomainind++) {
    candidatehead->start=loctransstartdomainind;
    for(loctransstopdomainind=transstoplowbound;
        loctransstopdomainind <
          MIN((transstoplowbound+TRANSUPSTREAMLIM+TRANSDOWNSTREAMLIM+1),
              dat4mfv->locuslength);
        loctransstopdomainind++) {
      candidatetail->stop=loctransstopdomainind;

      /* Score and update currmax as necessary */
      scoreisoform(candidatehead,dat4mfv,classifiercode,
        classifierparms,&classclown,&candidatescore);
      if(!(candidatescore < *currmaxscore)) {
        currmaxhead->start=candidatehead->start;
        currmaxtail->stop=candidatetail->stop;
        *currmaxscore=candidatescore;
        *isoformgoodp=classclown;
      }
    }
  }

  return;
} /* end findtranslationstartstop */

/* Predicate to report if isoform is already recorded in solutions */
static int solutionknownp(
  exoncoorsT *isoform,    /* candidate to compare            */
  exoncoorsT **solutions, /* solutions to compare against    */
  int structurect         /* count of solutions in solutions */
) {
  exoncoorsT
    *candidate=NULL, /* Candidate structure for recording      */
    *recorded=NULL;  /* Structure that was previously recorded */
  int sofarsogood,   /* bool used in comparison process        */
      i;             /* iterator variable                      */

  for(i=0;i<structurect;++i) {
    /* Compare solutions[i] and isoform */
    candidate=isoform;
    recorded=solutions[i];

    sofarsogood=TRUE;
    while(candidate!=NULL&&recorded!=NULL) {
      if(candidate->start!=recorded->start ||
         candidate->stop!=recorded->stop) {
        /* candidate and recorded differ */
        sofarsogood=FALSE;
        break;
      }
      else {
        candidate=candidate->nextelt;
        recorded=recorded->nextelt;
      }
    }
    if(sofarsogood==TRUE&&candidate==NULL&&recorded==NULL)
      return(TRUE); /* Solution has been seen previously */
  } /* end for */

  /* Solution is novel, return FALSE */
  return(FALSE);
} /* end solutionknownp */

/* Free memory in the linked list representing an expansion range */
static void freeexpansionrange(exonterminusrangeT *range) {
  exonterminusrangeT
    *current,*next; /* Pointers for traversing linked list */

  current=range;
  while(current->nextelt!=NULL) {
    next=current->nextelt;
    free(current);
    current=next;
  }
  free(current);
  range=NULL;

  return;
} /* end freeexpansionrange */

/* Free memory in the linked list representing an isoform */
static void freeisoform(exoncoorsT *isoform) {
  exoncoorsT
    *current,*next; /* Pointers for traversing linked list */

  current=isoform;
  while(current->nextelt!=NULL) {
    next=current->nextelt;
    free(current);
    current=next;
  }
  free(current);
  isoform=NULL;

  return;
} /* end freeisoform */

/* Reports all putative transcript isoforms predicted by PASIF *
 * using 1-based coordinates relative to locus.                *
 * (All processing by PASIF up to this point has been 0-based) */
static void dumppredictions(
  exoncoorsT **solutions, /* Set of putative transcript isoforms */
  double *solscores,      /* Corresponding scores for isoforms   */
  int isoformct,          /* Number of isoforms predicted        */
  char *locusid,          /* Identification for genomic sequence */
  int printscorep         /* Should scores be printed?           */
) {
  double keyscore;          /* for sorting by score           */
  int i,j;                  /* iterator variables             */
  exoncoorsT *current=NULL; /* Pointer to elements in list of *
                             * exon coordinates.              */

  /* We want to print isoform predictions in increasing scored order */
  if(!isoformct) /* Nothing was satisfactory enough to report! :( */
    ;
  else if(printscorep==TRUE)
    /* Insert sort of solutions, if solscores are "relevant". */
    for(j=1;j<isoformct;++j) {
      keyscore=solscores[j];
      current=solutions[j];
      for(i=j-1;i>=0&&solscores[i]<keyscore;--i) {
        solscores[i+1]=solscores[i];
        solutions[i+1]=solutions[i];
      }
      solscores[i+1]=keyscore;
      solutions[i+1]=current;
    }

  /* Because locusid is assuredly prefixed using '>'  *
   * (FASTA format) and this symbol would need to be  *
   * escaped using an XML entity, we will just ignore *
   * it in the output.                                */
  fprintf(XMLOUTSTREAM,"  <locus id=\"%s\">\n",++locusid);
  if(!isoformct)
    fprintf(XMLOUTSTREAM,"    <none />\n");
  else
    for(i=0;i<isoformct;++i) {
      current=solutions[i];
      if(printscorep==TRUE)
        fprintf(XMLOUTSTREAM,"    <isoform serial=\"%i\" score=\"%.4e\">\n",
          i+1,solscores[i]);
      else
        fprintf(XMLOUTSTREAM,"    <isoform serial=\"%i\">\n",i+1);
      while(current->nextelt != NULL) {
        fprintf(XMLOUTSTREAM,"      <exon start=\"%i\" stop=\"%i\"/>\n",
          current->start+1,current->stop+1);
        current=current->nextelt;
      }
      fprintf(XMLOUTSTREAM,"      <exon start=\"%i\" stop=\"%i\"/>\n",
        current->start+1,current->stop+1);
      fprintf(XMLOUTSTREAM,"    </isoform>\n");
    }
  fprintf(XMLOUTSTREAM,"  </locus>\n");

  return;
} /* end dumppredictions */
