/* splice_utils.c
 * Last modified: 16 June 2007
 *
 * In this file, the following functions were originally written by
 * Gordon Gremme for use with the GenomeThreader package, and modified
 * for use in this application by Michael E Sparks:
 *   evalsplicesiteprobformodel
 *   evaldonorprob
 *   evalacceptorprob
 *
 * These functions have the following copyright:
 *
 *   Copyright (c) 2003, 2005, 2006 Gordon Gremme <gremme@zbh.uni-hamburg.de>
 *   Copyright (c) 2006,2007 Michael E Sparks (mespar1@iastate.edu)
 *   All rights reserved.
 *
 * All other functions were written by Michael E Sparks, and exhibit
 * the following copyright:
 *
 *   Copyright (C) 2005,2006,2007 Michael E Sparks (mespar1@iastate.edu)
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notices and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "errors.h"
#include "immpractical.h"
#include "pasif_utils.h"
#include "splice_utils.h"

/* Function prototypes *******************************************************/

/* Function to return posterior probability that the given *
 * splice site is a true splice site, in any phase.        *
 * Extended to compute Bayes Factors and c-values by MES,  *
 * among other modifications.                              */
static void evalsplicesiteprobformodel(
  double *prob,         /* Records posterior probability     */
  double *bfac,         /* Records Bayes Factor              */
  double *cval,         /* Records BF critical value         */
  int donorsite,        /* TRUE for don site, FALSE for acc  */
  int *genomicsequence, /* int translation of locus          */
  int genomicfrompos,   /* index of first accessible residue */
  int genomictopos,     /* index of last accessible residue  */
  int genpos,           /* position of splice site in locus: *
                         * if donor, it indexes the first    *
                         * base of the dinucleotide; if acc, *
                         * the last.                         */
  Bssmmodel *bssmmodel  /* model parameterization            */
);

/* genpos indexes first base of donor dinucleotide. */
static double evaldonorprob(int *genomicsequence,
                            int genomicfrompos,
                            int genomictopos,
                            int genpos,
                            Bssmparm *bssmparm,
                            double *bfac,
                            double *cval);

/* genpos indexes last base of acceptor dinucleotide */
static double evalacceptorprob(int *genomicsequence,
                               int genomicfrompos,
                               int genomictopos,
                               int genpos,
                               Bssmparm *bssmparm,
                               double *bfac,
                               double *cval);

/* Calculates the rho-value of a splice site, assuming it is either *
 * a donor (donorsite == TRUE) or acceptor (donorsite == FALSE).    */
static double calc_rho(
  int *genomicsequence,int genomicfrompos,int genomictopos,
  int genpos,int donorsite,double *donscores,double *accscores,
  int *donp,int *accp);

/* Function definitions ******************************************************/

/* Function to return posterior probability that the given *
 * splice site is a true splice site, in any phase.        *
 * Extended to compute Bayes Factors and c-values by MES,  *
 * among other modifications.                              */
static void evalsplicesiteprobformodel(
  double *prob,         /* Records posterior probability     */
  double *bfac,         /* Records Bayes Factor              */
  double *cval,         /* Records BF critical value         */
  int donorsite,        /* TRUE for don site, FALSE for acc  */
  int *genomicsequence, /* int translation of locus          */
  int genomicfrompos,   /* index of first accessible residue */
  int genomictopos,     /* index of last accessible residue  */
  int genpos,           /* position of splice site in locus: *
                         * if donor, it indexes the first    *
                         * base of the dinucleotide; if acc, *
                         * the last.                         */
  Bssmmodel *bssmmodel  /* model parameterization            */
) {
  int pc, // previous char
      cc, // current char
      d, i, j;
  int startpos, endpos, termtype;
  double pval = DEFAULTSSPROB,
         Tv[3] = {0, 0, 0},
         Fv[4] = {0, 0, 0, 0},
         *priors=NULL,
         scaleupfac=0.0;

  // set start and endpos
  if (donorsite == TRUE)
  {
    termtype = 0; /* 0 indexes donor */
    startpos = genpos - bssmmodel->wsizedonorleft,
    endpos   = genpos + bssmmodel->wsizedonorright + 1;
  }
  else // acceptorsite
  {
    termtype = 1; /* 1 indexes acceptor */
    startpos = genpos - bssmmodel->wsizeacceptorleft - 1,
    endpos   = genpos + bssmmodel->wsizeacceptorright;
  }

  /* Note that genomicfrompos and genomictopos are valid index values */
  if(startpos<genomicfrompos||endpos>genomictopos) {
    /* Not possible to properly carry out the likelihood *
     * calculation, so return with default values.       */
    *prob=pval; /* Note that pval was set to DEFAULTSSPROB */
    *bfac=*cval=BFACCVALDNE;
    return;
  }

  /* Calculate splice site's likelihood */
  pc = genomicsequence[startpos];
  if (bssmmodel->hypothesisnum == HYPOTHESIS2)
  {
    Tv[0]=log((double)
            bssmmodel->hypotables.hypo2table[termtype][0][0][pc][0]);
    Fv[0]=log((double)
            bssmmodel->hypotables.hypo2table[termtype][1][0][pc][0]);
  }
  else
  {
    Tv[0]=log((double)
            bssmmodel->hypotables.hypo7table[termtype][0][0][pc][0]);
    Tv[1]=log((double)
            bssmmodel->hypotables.hypo7table[termtype][1][0][pc][0]);
    Tv[2]=log((double)
            bssmmodel->hypotables.hypo7table[termtype][2][0][pc][0]);
    Fv[0]=log((double)
            bssmmodel->hypotables.hypo7table[termtype][3][0][pc][0]);
    Fv[1]=log((double)
            bssmmodel->hypotables.hypo7table[termtype][4][0][pc][0]);
    Fv[2]=log((double)
            bssmmodel->hypotables.hypo7table[termtype][5][0][pc][0]);
    Fv[3]=log((double)
            bssmmodel->hypotables.hypo7table[termtype][6][0][pc][0]);
  }

  /* This step allows us to use                     *
   * 0 <= wsize[donor|acceptor]left <= MAXSPLICESIG */
  if (donorsite == TRUE)
    d = (int) MAXSPLICESIG - bssmmodel->wsizedonorleft;
  else
    d = (int) MAXSPLICESIG - bssmmodel->wsizeacceptorleft;

  for (i = (int) startpos + 1; i <= (int) endpos; i++)
  {
    j = d + (i - startpos); /* j tracks position in WAM w.r.t. i */
    cc = genomicsequence[i];
    if (bssmmodel->hypothesisnum == HYPOTHESIS2)
    {
      Tv[0]+=log((double)
               bssmmodel->hypotables.hypo2table[termtype][0][j][pc][cc]);
      Fv[0]+=log((double)
               bssmmodel->hypotables.hypo2table[termtype][1][j][pc][cc]);
    }
    else
    {
      Tv[0]+=log((double)
               bssmmodel->hypotables.hypo7table[termtype][0][j][pc][cc]);
      Tv[1]+=log((double)
               bssmmodel->hypotables.hypo7table[termtype][1][j][pc][cc]);
      Tv[2]+=log((double)
               bssmmodel->hypotables.hypo7table[termtype][2][j][pc][cc]);
      Fv[0]+=log((double)
               bssmmodel->hypotables.hypo7table[termtype][3][j][pc][cc]);
      Fv[1]+=log((double)
               bssmmodel->hypotables.hypo7table[termtype][4][j][pc][cc]);
      Fv[2]+=log((double)
               bssmmodel->hypotables.hypo7table[termtype][5][j][pc][cc]);
      Fv[3]+=log((double)
               bssmmodel->hypotables.hypo7table[termtype][6][j][pc][cc]);
    }
    pc = cc;
  }

  /* Compute splice site's posterior probability.                  *
   * Exponentiating logged joints COULD lead to buffer underflow   *
   * errors, so instead, we scale all logged likelihoods by some   *
   * reasonable constant, scaleupfac, prior to exponentiating. MES */
  if(bssmmodel->hypothesisnum==HYPOTHESIS2) {
    scaleupfac=(Tv[0]+Fv[0])/2.0;

    if((priors=(double*)malloc(sizeof(double)*2))==NULL)
      fatalerror("Err (evalsplicesiteprobformodel): Out of memory!\n");
    else
      for(i=0;i<2;++i) /* use an uninformative prior for now */
        priors[i]=1.0/2.0;

    pval=(priors[0]*exp(Tv[0]-scaleupfac))/
      (priors[0]*exp(Tv[0]-scaleupfac)+priors[1]*exp(Fv[0]-scaleupfac));
  }
  else {
    scaleupfac=(Tv[0]+Tv[1]+Tv[2]+Fv[0]+Fv[1]+Fv[2]+Fv[3])/7.0;

    if((priors=(double*)malloc(sizeof(double)*7))==NULL)
      fatalerror("Err (evalsplicesiteprobformodel): Out of memory!\n");
    else
      for(i=0;i<7;++i) /* use an uninformative prior */
        priors[i]=1.0/7.0;

    for(i=0,pval=0.0;i<3;++i)
      pval+=(priors[i]*exp(Tv[i]-scaleupfac))/
        (priors[0]*exp(Tv[0]-scaleupfac)+
         priors[1]*exp(Tv[1]-scaleupfac)+
         priors[2]*exp(Tv[2]-scaleupfac)+
         priors[3]*exp(Fv[0]-scaleupfac)+
         priors[4]*exp(Fv[1]-scaleupfac)+
         priors[5]*exp(Fv[2]-scaleupfac)+
         priors[6]*exp(Fv[3]-scaleupfac));
  }
  // confirm that pval is a valid probability
  if(pval<0.0||pval>1.0)
    fatalerror("Err (evalsplicesiteprobformodel): pval not valid!\n");

  /* Revert to DEFAULTSSPROB if necessary */
  if((2.0 * (pval - 0.5)) < DEFAULTSSPROB)
    pval = DEFAULTSSPROB;
  else
    pval = 2.0 * (pval - 0.5);
  // Set posterior probability of the splice site
  *prob = pval;

  /* Now, calculate Bayes Factor (BF)... */
  if(bssmmodel->hypothesisnum==HYPOTHESIS2)
    *bfac=exp(Tv[0]-scaleupfac)/exp(Fv[0]-scaleupfac);
  else
    *bfac=((priors[0]*exp(Tv[0]-scaleupfac)+
            priors[1]*exp(Tv[1]-scaleupfac)+
            priors[2]*exp(Tv[2]-scaleupfac))/
           (priors[0]+priors[1]+priors[2]))/
          ((priors[3]*exp(Fv[0]-scaleupfac)+
            priors[4]*exp(Fv[1]-scaleupfac)+
            priors[5]*exp(Fv[2]-scaleupfac)+
            priors[6]*exp(Fv[3]-scaleupfac))/
           (priors[3]+priors[4]+priors[5]+priors[6]));

  /* ...and the critical value, c. */
  *cval=2*log(*bfac);

  free(priors);
  return;
} /* end evalsplicesiteprobformodel */

/* genpos indexes first base of donor dinucleotide. */
static double evaldonorprob(int *genomicsequence,
                            int genomicfrompos,
                            int genomictopos,
                            int genpos,
                            Bssmparm *bssmparm,
                            double *bfac,
                            double *cval)
{
  double prob = DEFAULTSSPROB;

  if (genpos < genomictopos)
  {
    if (bssmparm->gtmodelset == TRUE &&
        genomicsequence[genpos] == 2 &&
        genomicsequence[genpos + 1] == 3)
    {
      evalsplicesiteprobformodel(&prob,
                                 bfac,
                                 cval,
                                 TRUE, // is donor site
                                 genomicsequence,
                                 genomicfrompos,
                                 genomictopos,
                                 genpos,
                                 &bssmparm->gtmodel);
    }
    else if (bssmparm->gcmodelset == TRUE &&
            genomicsequence[genpos] == 2 &&
            genomicsequence[genpos + 1] == 1)
    {
      evalsplicesiteprobformodel(&prob,
                                 bfac,
                                 cval,
                                 TRUE, // is donor site
                                 genomicsequence,
                                 genomicfrompos,
                                 genomictopos,
                                 genpos,
                                 &bssmparm->gcmodel);
    }
  }
  return(prob);
}

/* genpos indexes last base of acceptor dinucleotide */
static double evalacceptorprob(int *genomicsequence,
                               int genomicfrompos,
                               int genomictopos,
                               int genpos,
                               Bssmparm *bssmparm,
                               double *bfac,
                               double *cval)
{
  double prob = DEFAULTSSPROB;

  if (genpos > 0 &&
      genomicsequence[genpos - 1] == 0 &&
      genomicsequence[genpos] == 2)
  {
    evalsplicesiteprobformodel(&prob,
                               bfac,
                               cval,
                               FALSE, // is acceptor site
                               genomicsequence,
                               genomicfrompos,
                               genomictopos,
                               genpos,
                               &bssmparm->gtmodel);
  }
  return(prob);
}

/* Calculates the rho-value of a splice site, assuming it is either *
 * a donor (donorsite==TRUE) or acceptor (donorsite==FALSE).        */
static double calc_rho(
  int *genomicsequence,int genomicfrompos,int genomictopos,
  int genpos,int donorsite,double *donscores,double *accscores,
  int *donp,int *accp
) {
  /* Variable names here roughly follow those *
   * from Fig.1 of Brendel and Kleffe 1998    */
  int d,dcom, /* donor sites being scored                   */
      a,acom, /* acceptor sites being scored                */
      r,rcom, /* competing donor sites, downstream of d     */
      l,lcom, /* acceptors competing with a, upstream of d  */
      start,locstart, /* start bound of search              */
      stop,locstop,   /* stop bound of search               */
      pushpairp, /* flag for when a don/acc pair was found  */
      pairct=0,  /* Total number of competing don/acc pairs */
      i;         /* iterator variable                       */
  double rho=0.0,        /* Value to compute and return     */
         *pvallist=NULL, /* stores don/acc P-values         */
         bestscore;      /* used for comparing scores       */

  if(donorsite==TRUE&&donp[genpos]==TRUE) {
    d=genpos; /* First base of donor dinucleotide */

    /* Now, we need to compute the extent of genomicsequence *
     * that can be searched in for a cognate acceptor to d.  *
     * First, the start bound:                               */
    if((start=d+lengthlimits[MININTRONLENX]-1)>genomictopos)
      /* Spilled off edge of the world, so no acceptor match possible. */
      return(rho);
    /* Next, the stopping boundary: */
    if((stop=d+lengthlimits[MAXINTRONLENX]-1)>genomictopos)
      stop=genomictopos;
    /* check for competing, downstream donor r */
    for(r=start+1;r<stop-lengthlimits[MININTRONLENX]+2;++r)
      if(donp[r]==TRUE&&donscores[r]>donscores[d]) {
        stop=r+lengthlimits[MININTRONLENX]-2;
        break;
      }
    /* Find cognate acceptor a */
    for(i=start,bestscore=-0.1,a=-1;i<=stop;++i)
      if(accp[i]==TRUE&&accscores[i]>bestscore) {
        bestscore=accscores[i];
        a=i;
      }
    if(a==-1) /* no acc found (see init in above for-loop) */
      return(rho);
    else
      pairct=1;
    /* We now have the primary don/acc pair for scoring */
    if((pvallist=(double*)malloc(sizeof(double)*2))==NULL)
      fatalerror("Err (calc_rho): Out of memory!\n");
    else {
      pvallist[0]=donscores[d];
      pvallist[1]=accscores[a];
    }

    /* Now identify all local, competing don/acc pairs. *
     * First, the start bound:                          */
    if((start=a-lengthlimits[MAXINTRONLENX]+1)<genomicfrompos)
      start=genomicfrompos;
    /* check for competing acceptor l, upstream of d */
    for(l=d;l>start+lengthlimits[MININTRONLENX]-2;--l)
      if(accp[l]==TRUE&&accscores[l]>accscores[a]) {
        start=l-lengthlimits[MININTRONLENX]+2;
        break;
      }
    /* Next, the stopping boundary: */
    stop=a-1; /* Since an acceptor a exists, a possible donor   *
               * at a-1 must, too, i.e., we are not outside the *
               * limits of the genomic locus at this position.  */
    for(dcom=start;dcom<=stop;++dcom) {
      if(dcom==d) /* already paired with a */
        continue;

      if(donp[dcom]==TRUE) { /* find cognate acceptor, if it exists */
        pushpairp=FALSE;
        if(dcom<d) {
          locstart=dcom+lengthlimits[MININTRONLENX]-1;
          if((locstop=dcom+lengthlimits[MAXINTRONLENX]-1)>a)
            locstop=a;
          for(i=locstart,bestscore=-0.1,acom=-1;i<=locstop;++i)
            if(accp[i]==TRUE&&accscores[i]>bestscore) {
              bestscore=accscores[i];
              acom=i;
            }
          if(acom==-1) /* Should at least be able to pair with a! */
            fatalerror("Err (calc_rho): Missed acceptor a?!!\n");
          else
            pushpairp=TRUE;
        } 
        else { /* in interval [d+1,a-1], the strategy is *
                * basically that for finding a w.r.t. d  */
          /* Now, we need to compute the extent of genomicsequence   *
           * that can be searched in for a cognate acceptor to dcom. *
           * First, the start bound:                                 */
          if((locstart=dcom+lengthlimits[MININTRONLENX]-1)>genomictopos)
            /* Spilled off edge of the world, so no acceptor match *
             * is possible for this dcom, or any downstream.       */
            break;
          /* Next, the stopping boundary: */
          if((locstop=dcom+lengthlimits[MAXINTRONLENX]-1)>genomictopos)
            locstop=genomictopos;
          /* check for competing, downstream donor rcom */
          for(rcom=locstart+1;
              rcom<locstop-lengthlimits[MININTRONLENX]+2;
              ++rcom)
            if(donp[rcom]==TRUE&&donscores[rcom]>donscores[dcom]) {
              locstop=rcom+lengthlimits[MININTRONLENX]-2;
              break;
            }
          /* Find cognate acceptor acom for dcom */
          for(i=locstart,bestscore=-0.1,acom=-1;i<=locstop;++i)
            if(accp[i]==TRUE&&accscores[i]>bestscore) {
              bestscore=accscores[i];
              acom=i;
            }
          if(acom==-1) /* no acc found (see init in above for-loop) */
            continue;
          else
            pushpairp=TRUE;
        } 

        if(pushpairp==TRUE) { /* push competing pair onto pvallist */
          ++pairct;
          if((pvallist=(double*)
                realloc(pvallist,(pairct<<1)*sizeof(double)))==NULL)
            fatalerror("Err (calc_rho): Out of memory!\n");
          else {
            pvallist[(pairct-1)<<1]=donscores[dcom];
            pvallist[(pairct<<1)-1]=accscores[acom];
          }
        } 
      } /* end donp==true if */
    } /* end dcom for */
  } /* end of if for calculating rho under donor identity */
  else if(donorsite==FALSE&&accp[genpos]==TRUE) {
    a=genpos; /* last base of acceptor dinucleotide */

    /* Now, we need to compute the extent of genomicsequence *
     * that can be searched in for a cognate donor d for a.  *
     * First, the stopping boundary:                         */
    if((stop=a-lengthlimits[MININTRONLENX]+1)<genomicfrompos)
      /* Spilled off edge of the world, so no donor match possible. */
      return(rho);
    /* Next, the start bound: */
    if((start=a-lengthlimits[MAXINTRONLENX]+1)<genomicfrompos)
      start=genomicfrompos;
    /* check for competing, upstream acceptor l */
    for(l=stop-1;l>start+lengthlimits[MININTRONLENX]-2;--l)
      if(accp[l]==TRUE&&accscores[l]>accscores[a]) {
        start=l-lengthlimits[MININTRONLENX]+2;
        break;
      }
    /* Find cognate donor d for a */
    for(i=start,bestscore=-0.1,d=-1;i<=stop;++i)
      if(donp[i]==TRUE&&donscores[i]>bestscore) {
        bestscore=donscores[i];
        d=i;
      }
    if(d==-1) /* no don found (see init in above for-loop) */
      return(rho);
    else
      pairct=1;
    /* We now have the primary don/acc pair for scoring */
    if((pvallist=(double*)malloc(sizeof(double)*2))==NULL)
      fatalerror("Err (calc_rho): Out of memory!\n");
    else {
      pvallist[0]=donscores[d];
      pvallist[1]=accscores[a];
    }

    /* Now identify all local, competing don/acc pairs. *
     * First, the starting boundary:                    */
    start=d+1;
    /* Next, the stopping bound: */
    if((stop=d+lengthlimits[MAXINTRONLENX]-1)>genomictopos)
      stop=genomictopos;
    /* check for competing donor r, downstream of a */
    for(r=a+1;r<stop-lengthlimits[MININTRONLENX]+2;++r)
      if(donp[r]==TRUE&&donscores[r]>donscores[d]) {
        stop=r+lengthlimits[MININTRONLENX]-2;
        break;
      }

    for(acom=start;acom<=stop;++acom) {
      if(acom==a) /* Already pairs with d */
        continue;

      if(accp[acom]==TRUE) { /* find cognate donor, if exists */
        pushpairp=FALSE;

        if(acom>a) {
          if((locstart=acom-lengthlimits[MAXINTRONLENX]+1)<d)
            locstart=d; /* acom can't find a better partner than d upstream *
                         * of d, or a would have found it, already.         */
          locstop=acom-lengthlimits[MININTRONLENX]+1;
          for(i=locstart,bestscore=-0.1,dcom=-1;i<=locstop;++i)
            if(donp[i]==TRUE&&donscores[i]>bestscore) {
              bestscore=donscores[i];
              dcom=i;
            }
          if(dcom==-1) /* Should at least be able to pair with d! */
            fatalerror("Err (calc_rho): Missed donor d?!!\n");
          else
            pushpairp=TRUE;
        } 
        else { /* in interval [d+1,a-1], the strategy is *
                * basically that for finding d w.r.t. a  */
          /* Now, we need to compute the extent of genomicsequence      *
           * that can be searched in for a cognate donor dcom for acom. *
           * First, the start bound:                                    */
          if((locstop=acom-lengthlimits[MININTRONLENX]+1)<genomicfrompos)
            /* Spilled off edge of the world, so no donor match *
             * is possible for this acom, nor any upstream.     */
            break;
          /* Next, the start bound: */
          if((locstart=acom-lengthlimits[MAXINTRONLENX]+1)<genomicfrompos)
            locstart=genomicfrompos;
          /* check for competing, upstream acceptor lcom */
          for(lcom=locstop-1;
              lcom>locstart+lengthlimits[MININTRONLENX]-2;
              --lcom)
            if(accp[lcom]==TRUE&&accscores[lcom]>accscores[acom]) {
              locstart=lcom-lengthlimits[MININTRONLENX]+2;
              break;
            }
          /* Find cognate donor dcom for acom */
          for(i=locstart,bestscore=-0.1,dcom=-1;i<=locstop;++i)
            if(donp[i]==TRUE&&donscores[i]>bestscore) {
              bestscore=donscores[i];
              dcom=i;
            }
          if(dcom==-1) /* no don found (see init in above for-loop) */
            continue;
          else
            pushpairp=TRUE;
        } 

        if(pushpairp==TRUE) { /* push competing pair onto list */
          ++pairct;
          if((pvallist=(double*)
               realloc(pvallist,(pairct<<1)*sizeof(double)))==NULL)
            fatalerror("Err (calc_rho): Out of memory!\n");
          else {
            pvallist[(pairct-1)<<1]=donscores[dcom];
            pvallist[(pairct<<1)-1]=accscores[acom];
          }
        } 
      } /* end accp==true if */
    } /* end acom for */
  } /* end of if for calculating rho under acceptor identity */
  else
    return(rho);

  /* Finally, we can compute rho */
  for(i=0,rho=0.0;i<pairct<<1;i+=2)
    rho+=pvallist[i]*pvallist[i+1];
  rho=pow(rho,-1.0)*pow(pvallist[0]*pvallist[1],2);

  free(pvallist);
  return(rho);
} /* end calc_rho */

/* Function to decorate the locus with posterior probabilities        *
 * of each site as being an acceptor or donor dinucleotide terminus.  *
 * Position i in each array will store the probability that the       *
 * nucleotide at position i is the first base of a donor dinucleotide *
 * or the last base of an acceptor dinucleotide.                      *
 * Additionally, rho-values as described in Brendel and Kleffe's 1998 *
 * NAR paper are computed for each position in the sequence, under    *
 * identities of the site in question being a donor or acceptor.      */
void analyze_locus_splice_sites(
  Bssmparm *bssmparms,int len,int *locus,
  double *donscores,double *accscores,
  double *rho_vals_don,double *rho_vals_acc
) {
  int i,          /* iterator variable                     */
      *donp=NULL, /* Can site be a donor?                  */
      *accp=NULL; /* Can site be an acceptor?              */
  double bfac,    /* Stores Bayes Factor for a given site  */
         cval;    /* Stores corresponding c-value for bfac */

  if((donp=(int*)malloc(sizeof(int)*len))==NULL ||
     (accp=(int*)malloc(sizeof(int)*len))==NULL)
    fatalerror("Err (analyze_locus_splice_sites): Out of memory!\n");
  else
    for(i=0;i<len;++i)
      donp[i]=accp[i]=FALSE;

  /* Posterior probabilities will duplicate the role of the *
   * original logitlinear P-values of Brendel and Kleffe,   *
   * 1998.  Because rho-values depend on these data, they   *
   * will be computed first.  Also, not every site need be  *
   * considered a donor or acceptor.  To have a SNOBOL's    *
   * chance, their c-value must meet or exceed MINCVAL.     */
  for(i=0;i<len;++i) {
    cval=MINCVAL-1;
    donscores[i]=evaldonorprob(locus,0,len-1,i,bssmparms,&bfac,&cval);
    if(!(cval<MINCVAL))
      donp[i]=TRUE;

    cval=MINCVAL-1;
    accscores[i]=evalacceptorprob(locus,0,len-1,i,bssmparms,&bfac,&cval);
    if(!(cval<MINCVAL))
      accp[i]=TRUE;
  }

  for(i=0;i<len;++i) {
    rho_vals_don[i]=
      calc_rho(locus,0,len-1,i,TRUE,donscores,accscores,donp,accp);
    rho_vals_acc[i]=
      calc_rho(locus,0,len-1,i,FALSE,donscores,accscores,donp,accp);
  }

  free(donp);
  free(accp);

  return;
} /* end analyze_locus_splice_sites */

/* Function to import a trained Bssmparm object *
 * (stored in a binary file) into main memory.  */
void import_Bssmparm(char *filename,Bssmparm *probs)
{
  FILE *fptr=NULL; /* For connecting to binary input file */

  /* Open parameter file */
  if((fptr=fopen(filename,"rb"))==NULL) {
    (void)snprintf(errbuff,MAXERRLEN,"Err (import_Bssmparm): \
Can't open %s for binary reading!\n",filename);
    fatalerror(errbuff);
  }
  /* Read in the Bssmparm object */
  else if(fread(probs,sizeof(Bssmparm),1,fptr)!=1) {
    (void)snprintf(errbuff,MAXERRLEN,"Err (import_Bssmparm): \
Error reading from %s!\n",filename);
    fatalerror(errbuff);
  }
  else
    (void)fclose(fptr);

  return;
} /* end import_Bssmparm */
